/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <string.h>
#include <stdlib.h>

#include "render.h"
#include "GB.h"
#include "cpu.h"
#include "debug.h"
#include "mainloop.h"
#include "gfx.h"
#include "mem.h"

int gb_system::gfx_bit_count = 16;

gb_system::gb_system():
skip_frame(0), 
off_counter(0), 
serialbits(0), 
IWait(0), 
STAT_wait(0),
draw_wait(0), 
frameskip_counter(0), 
mode3_cyc(0),
wy_set(-1), 
IME(0),
CPUHalt(0), 
gb_speed(0), 
int_line(-1), 
timer_freq(CYCLES_TIMER_MODE0),
cycles_DIV(CYCLES_DIV), 
cycles_timer(CYCLES_TIMER_MODE0), 
cycles_LCD(CYCLES_LCD_MODE2),
cycles_serial(0), 
cycles_sound(0), 
EI_count(0), 
timeron(0), 
LCDon(1),
tile_map(0), 
tile_pattern(0), 
win_tile_map(0), 
sprite_number(0), 
windowline(-1),
memory(NULL),
cartRAM(NULL),
VRAM(NULL),
WRAM(NULL),
cartridge(NULL),
hdma_source(0x0000),
hdma_destination(0x8000),
hdma_bytes(0),
hdma_on(0),
gdma_rest(0),
gfx_buffer1(NULL),
gfx_buffer2(NULL),
romloaded(false),
system_type(SYS_AUTO),
rom(NULL),
gbc_mode(0),
A(1),
F(0xB0),
ZFLAG(1),
HFLAG(1),
CFLAG(1),
NFLAG(1),
flags(0x5100),
CYCLES_SERIAL(CYCLES_SERIAL_GB),
MBC1memorymodel(0),
RAMenable(0),
rom_bank(1),
ram_bank(0),
wram_bank(1),
vram_bank(0),
MBChi(0),
MBClo(0),
RTCIO(0)
{
   button_pressed[B_LEFT]=button_pressed[B_RIGHT]=button_pressed[B_DOWN]=button_pressed[B_UP]=1;
   button_pressed[B_START]=button_pressed[B_SELECT]=button_pressed[B_A]=button_pressed[B_B]=1;
 
   BC.W = 0x0013;
   DE.W = 0x00D8;
   HL.W = 0x014D;
   SP.W = 0xFFFE;
   PC.W = 0x0100;
   
   BGP[0] = 0;
   BGP[1] = 3;
   BGP[2] = 3;
   BGP[3] = 3;         
   
   OBP0[0] = 3;
   OBP0[1] = 3;
   OBP0[2] = 3;
   OBP0[3] = 3;     
   
   OBP1[0] = 3;    
   OBP1[1] = 3;
   OBP1[2] = 3;
   OBP1[3] = 3;         
}

bool gb_system::init()
{
   memory = new byte[65536];
   if(!memory) 
      return false;
      
   cartRAM = new byte[128*1024];
   if(!cartRAM) 
      return false;
   memset(cartRAM,0xFF,128*1024);
   
   VRAM = new byte[16384];
   if(!VRAM)
      return false;
      
   WRAM = new byte[32768];
   if(!WRAM)
      return false;
      
   rom = new GBrom;
   if(!rom)
      return false;
           
   return true;
}

bool gb_system::init_gfx()
{
   if(gfx_bit_count == 16)
   {
      gfx_buffer1 = new WORD[160*144];
      gfx_buffer2 = new WORD[160*144];
   } else
   {
      gfx_buffer1 = new DWORD[160*144];
      gfx_buffer2 = new DWORD[160*144];   
   }      
   if(gfx_buffer1 == NULL || gfx_buffer2 == NULL)
      return false;   
      
   gfx_buffer = gfx_buffer1;
   gfx_buffer_old = gfx_buffer2;
      
   return true;
}

gb_system::~gb_system() 
{
   if(memory != NULL)
   {
      delete [] memory;
      memory = NULL;
   }
   if(cartRAM != NULL)
   {
      delete [] cartRAM;
      cartRAM = NULL;
   }
   if(VRAM != NULL)
   {
      delete [] VRAM;
      VRAM = NULL;
   }
   if(WRAM != NULL)
   {
      delete [] WRAM;
      WRAM = NULL;
   }      
   if(gfx_buffer1 != NULL)
   {
      if(gfx_bit_count == 16)
         delete [] (WORD*)gfx_buffer1;
      else
         delete [] (DWORD*)gfx_buffer1;
      gfx_buffer1 = NULL;
   }      
   if(gfx_buffer2 != NULL)
   {
      if(gfx_bit_count == 16)
         delete [] (WORD*)gfx_buffer2;
      else
         delete [] (DWORD*)gfx_buffer2;
      gfx_buffer2 = NULL;
   }        
   if(cartridge != NULL)
   {
      delete [] cartridge;
      cartridge = NULL;
   }
   if(rom != NULL)
   {
      delete rom;
      rom = NULL;
   }
}

void gb_system::reset()
{
   int old_sgb_mode = sgb_mode;
   int old_gbc_mode = gbc_mode;
   
   //change mode according to user selection
   if(system_type == SYS_GBP || system_type == SYS_GB)
      sgb_mode = gbc_mode = 0;
   else
   if(system_type == SYS_GBC || system_type == SYS_GBA)
   {
      gbc_mode = 1;
      sgb_mode = 0;
      if(sgb_border_option == SGB_BORDER && rom->SGB)
      {
         sgb_mode = 1;
         gbc_mode = 1;
      } else
      if(sgb_border_option == SGB_INITB && rom->SGB)
      {
         sgb_mode = 1;
         gbc_mode = 0;               
      }
   }
   else
   if((system_type == SYS_SGB || system_type == SYS_SGB2) || (sgb_border_option && rom->CGB && rom->SGB)) 
   {       
      sgb_mode = 1;
      gbc_mode = 0;
      if(rom->CGB && sgb_border_option==SGB_BORDER)
         gbc_mode = 1;
   }
   else
   {
      sgb_mode = gbc_mode = 0;
      
      if(rom->CGB)
         gbc_mode = 1;

      if(rom->SGB && !gbc_mode)
         sgb_mode = 1;
   }
   
   if(multiple_gb && sgb_mode) // don't allow SGB mode
   {
      sgb_mode = 0;
      if(rom->CGB) gbc_mode = 1;
   }
         
   if(sgb_mode)
   {
      fill_gfx_buffers(0UL);             
   } else   
   if(gbc_mode)
   {
      if(gfx_bit_count==16)
         fill_gfx_buffers(gfx_pal16[0x7fff]|(gfx_pal16[0x7fff]<<16)); 
      else
         fill_gfx_buffers(gfx_pal32[0x7fff]);               
   } else   
   {
      if(gfx_bit_count==16)
         fill_gfx_buffers(gfx_pal16[GB->GBC_BGP[0]]|(gfx_pal16[GB->GBC_BGP[0]]<<16)); 
      else
         fill_gfx_buffers(gfx_pal32[GB->GBC_BGP[0]]);                 
   }
   
   if(gbc_mode)
      draw_line_tile = draw_line_tile_GBC;
   else
      draw_line_tile = draw_line_tile_DMG;
         
   if(gbc_real_colors && ((old_gbc_mode^gbc_mode) || system_type == SYS_GBC || system_type == SYS_GBA || system_type == SYS_AUTO))
      mix_gbc_colors();
   
   //resize the window
   if(!old_sgb_mode && sgb_mode)
   {
      resize_window(256,224);
   } else 
   if(old_sgb_mode && !sgb_mode)
   {
      resize_window(160,144);                   
   }
   
   button_pressed[B_LEFT]=button_pressed[B_RIGHT]=button_pressed[B_DOWN]=button_pressed[B_UP]=1;
   button_pressed[B_START]=button_pressed[B_SELECT]=button_pressed[B_A]=button_pressed[B_B]=1;
 
   IWait = 0;
   STAT_wait = 0;
   draw_wait = 0;
      
   sprite_number = 0;
   off_counter = 0;
   serialbits = 0;
   
   timeron = 0;
   LCDon = 1;
   
   windowline = -1;
   
   hdma_on = 0;
   hdma_source = 0;
   hdma_destination = 0x8000;
   hdma_bytes = 0;
   gdma_rest = 0;
   
   mem_reset();
   
   cpu_reset();
}

void gb_system::cpu_reset()
{   
   A = 0x01;
   F = 0xB0;
   ZFLAG = HFLAG = CFLAG = 1;
   NFLAG = 0;
   flags = 0x5100;
      
   BC.W = 0x0013;
   DE.W = 0x00D8;
   HL.W = 0x014D;
   SP.W = 0xFFFE;
   PC.W = 0x0100;
   
   CPUHalt = 0;
   IME = 0;
   EI_count = 0; 
   int_line = -1;
   
   cycles_DIV = CYCLES_DIV;
   cycles_timer = CYCLES_TIMER_MODE0;
   cycles_LCD = CYCLES_LCD_MODE2;
   cycles_serial = 0;  
   gb_speed = 0;
   CYCLES_SERIAL = CYCLES_SERIAL_GB;
   
   if(system_type == SYS_GBA)
      BC.W = 0x0113;
   else
   if(system_type == SYS_GBP || system_type == SYS_SGB2)
      A = 0xFF;
   
   if(gbc_mode)
      A = 0x11;
}

void gb_system::mem_reset()
{
   memset(memory+0x8000,0x00,0x1FFF);
   memset(memory+0xFE00,0x00,0xA0);
   
   if(rom->bankType==MBC2)   
      for(int a=0xa000;a<0xc000;++a)
         memory[a] = 0x0F;
   else
      for(int a=0xa000;a<0xc000;++a)
         memory[a] = 0xFF;
         
   // Beat Mania 2
   memory[0xbccc] = 0xC9; // set RET opcode where Beat Mania 2 jumps   

   // Joust & Defender (U) [C][t1]  &  Fire Gear (Bung)
   for(int a=0xc000;a<0xcfff;++a)
      memory[a] = 0x00;   
   memory[0xc100] = 0xff;// fix for Minesweeper for 'Windows'
         
   mem_map[0x0] = &cartridge[0x0000];
   mem_map[0x1] = &cartridge[0x1000];
   mem_map[0x2] = &cartridge[0x2000];
   mem_map[0x3] = &cartridge[0x3000];
   mem_map[0x4] = &cartridge[0x4000];
   mem_map[0x5] = &cartridge[0x5000];
   mem_map[0x6] = &cartridge[0x6000];
   mem_map[0x7] = &cartridge[0x7000];

   if(gbc_mode)
   {
      mem_map[0x8] = &VRAM[0x0000];
      mem_map[0x9] = &VRAM[0x1000];
      mem_map[0xA] = &memory[0xA000];
      mem_map[0xB] = &memory[0xB000];
      mem_map[0xC] = &memory[0xC000];
      mem_map[0xD] = &WRAM[0x1000];
      mem_map[0xE] = &memory[0xE000];
      mem_map[0xF] = &memory[0xF000];
   } else
   {
      mem_map[0x8] = &memory[0x8000];
      mem_map[0x9] = &memory[0x9000];
      mem_map[0xA] = &memory[0xA000];
      mem_map[0xB] = &memory[0xB000];
      mem_map[0xC] = &memory[0xC000];
      mem_map[0xD] = &memory[0xD000];
      mem_map[0xE] = &memory[0xE000];
      mem_map[0xF] = &memory[0xF000];
   }
   
   if(rom->RAMsize>2)
   {
      mem_map[0xA] = &cartRAM[0x0000];
      mem_map[0xB] = &cartRAM[0x1000];      
   }         
   
   memset(memory+0xFEA0,0xFF,0x60);

   memory[0xFF00] = 0xFF; // P1
   memory[0xFF01] = 0x00; // SB
   memory[0xFF02] = 0x00; // SC
   memory[0xFF04] = 0x00; // DIV
   memory[0xFF05] = 0x00; // TIMA
   memory[0xFF06] = 0x00; // TMA
   memory[0xFF07] = 0x00; // TAC
   memory[0xFF0F] = 0xE1; // IF
   memory[0xFF10] = 0x80; // NR10
   memory[0xFF11] = 0xBF; // NR11
   memory[0xFF12] = 0xF3; // NR12
   memory[0xFF13] = 0xFF; // NR13   
   memory[0xFF14] = 0xBF; // NR14
   memory[0xFF15] = 0xFF; // NR20   
   memory[0xFF16] = 0x3F; // NR21
   memory[0xFF17] = 0x00; // NR22
   memory[0xFF18] = 0xFF; // NR23  
   memory[0xFF19] = 0xBF; // NR24
   memory[0xFF1A] = 0x7F; // NR30
   memory[0xFF1B] = 0xFF; // NR31
   memory[0xFF1C] = 0x9F; // NR32
   memory[0xFF1D] = 0xFF; // NR33  
   memory[0xFF1E] = 0xBF; // NR33
   memory[0xFF1F] = 0xFF; // NR40   
   memory[0xFF20] = 0xFF; // NR41
   memory[0xFF21] = 0x00; // NR42
   memory[0xFF22] = 0x00; // NR43
   memory[0xFF23] = 0xBF; // NR44
   memory[0xFF24] = 0x77; // NR50
   memory[0xFF25] = 0xF3; // NR51
   memory[0xFF26] = 0xF1; // NR52   
   memory[0xFF40] = 0x91; // LCDC
   memory[0xFF41] = 0x81; // STAT
   memory[0xFF42] = 0x00; // SCY
   memory[0xFF43] = 0x00; // SCX
   memory[0xFF44] = 0x00; // LY
   memory[0xFF45] = 0x00; // LYC
   memory[0xFF46] = 0x00; // DMA
   memory[0xFF47] = 0xFC; // BGP  
   memory[0xFF48] = 0xFF; // OBP0
   memory[0xFF49] = 0xFF; // OBP1  
   memory[0xFF4A] = 0x00; // WY
   memory[0xFF4B] = 0x00; // WX
   memory[0xFF4F] = 0xFE; // VBK
   memory[0xFF4D] = 0x00; // KEY1
   memory[0xFF51] = 0x00; // HDMA1
   memory[0xFF52] = 0x00; // HDMA2
   memory[0xFF53] = 0x00; // HDMA3
   memory[0xFF54] = 0x00; // HDMA4
   memory[0xFF55] = 0x00; // HDMA5
   memory[0xFF68] = 0x00; // BCPS
   memory[0xFF6A] = 0x00; // OCPS
   memory[0xFF69] = 0x00; // BCPD
   memory[0xFF6B] = 0x00; // OCPD
   memory[0xFF70] = 0x00; // SVBK
   memory[0xFFFF] = 0x00; // IE

   BGP[0] = 0;
   BGP[1] = 3;
   BGP[2] = 3;
   BGP[3] = 3;

   OBP0[0]=OBP0[1]=OBP0[2]=OBP0[3]=3;
   OBP1[0]=OBP1[1]=OBP1[2]=OBP1[3]=3;  
   
   MBC1memorymodel = 0;
   MBChi = 0;
   MBClo = 1;
   rom_bank = 1;
   ram_bank = 0;
   wram_bank = 1;
   vram_bank = 0;
   RTCIO = 0;
   
   if(gbc_mode)
   {
      if(rom->CGB)
      {
         hdma_source = 0x0000;
         hdma_destination = 0x8000;
         memory[0xFF44] = 0x90; // LY   
         memory[0xFF45] = 0x00; // LYC      
         memory[0xFF41] = 0x81; // STAT  
         memory[0xFF55] = 0xFF; // HDMA5
         memory[0xFF56] = 0x02; // RP
         memory[0xFF68] = 0xC0; // BCPS
         memory[0xFF6A] = 0xC0; // OCPS

         memory[0xFF6C] = 0xFE; // undocumented 
         memory[0xFF72] = 0x00; // undocumented  
         memory[0xFF73] = 0x00; // undocumented  
         memory[0xFF74] = 0x00; // undocumented   
         memory[0xFF75] = 0x8F; // undocumented                      
         memory[0xFF76] = 0x00; // undocumented
         memory[0xFF77] = 0x00; // undocumented      
         for(int i=0;i<32;)
         {
            GBC_BGP[i++] = 0x7FFF; 
            GBC_BGP[i++] = 0x7FFF;
            GBC_BGP[i++] = 0x7FFF;
            GBC_BGP[i++] = 0x7FFF;                   
         } 
           
         if(!strcmp(rom->name,"BUGS CC3 CRACK") || !strcmp(rom->name,"OS"))
            memcpy(GBC_BGP,GBC_DMGBG_palettes[0],sizeof(unsigned short)*4);
            
         memcpy(GBC_OBP,GBC_DMGBG_palettes[0],sizeof(unsigned short)*4);
      }
   }
   
   tile_map = ((memory[0xFF40] & 0x08)?0x1c00:0x1800);
   win_tile_map = ((memory[0xFF40] & 0x40)?0x1c00:0x1800);
   tile_pattern = ((memory[0xFF40] & 0x10)?0x0000:0x0800);

   if(gbc_mode && !rom->CGB)
   {
      memcpy(GBC_BGP,GBC_DMGBG_palettes[1],sizeof(unsigned short)*4);
      memcpy(GBC_OBP,GBC_DMGOBJ0_palettes[1],sizeof(unsigned short)*4);
      memcpy(GBC_OBP+4,GBC_DMGOBJ1_palettes[1],sizeof(unsigned short)*4);
   } else
   if(!rom->CGB || !gbc_mode)
   {
      if(color_option == COL_BW)
      {
         memcpy(GBC_BGP,GBC_DMGBG_palettes[0],sizeof(unsigned short)*4);
         memcpy(GBC_OBP,GBC_DMGOBJ0_palettes[0],sizeof(unsigned short)*4);
         memcpy(GBC_OBP+4,GBC_DMGOBJ1_palettes[0],sizeof(unsigned short)*4);   
      } else
      if(color_option == COL_LCDBROWN)
      {
         memcpy(GBC_BGP,LCD_palette_brown,sizeof(unsigned short)*4);
         memcpy(GBC_OBP,LCD_palette_brown,sizeof(unsigned short)*4);
         memcpy(GBC_OBP+4,LCD_palette_brown,sizeof(unsigned short)*4);
      } else
      if(color_option == COL_LCDGREEN)
      {
         memcpy(GBC_BGP,LCD_palette_green,sizeof(unsigned short)*4);
         memcpy(GBC_OBP,LCD_palette_green,sizeof(unsigned short)*4);
         memcpy(GBC_OBP+4,LCD_palette_green,sizeof(unsigned short)*4);
      } else
      if(color_option == COL_GRAY)
      {
         memcpy(GBC_BGP,GB_palette_gray,sizeof(unsigned short)*4);
         memcpy(GBC_OBP,GB_palette_gray,sizeof(unsigned short)*4);
         memcpy(GBC_OBP+4,GB_palette_gray,sizeof(unsigned short)*4);
      }
   }   
   
   memset(VRAM,0,16384);
}

void gb_system::do_hdma()
{
  copy_memory(hdma_destination,hdma_source,0x10);

  hdma_destination += 0x10;
  hdma_source += 0x10;
  
  memory[0xFF52] += 0x10;
  if(memory[0xFF52] == 0x00)
    ++memory[0xFF51];
  
  memory[0xFF54] += 0x10;
  if(memory[0xFF54] == 0x00)
    ++memory[0xFF53];
      
  hdma_bytes -= 0x10;
  --memory[0xFF55];

  if(memory[0xFF55] == 0xff)
    hdma_on = 0;
}

void gb_system::handle_interrupt()
{
   CPUHalt = IME = EI_count = 0;
   PUSH(PC.B.h); 
   PUSH(PC.B.l);       

   if(memory[0xFFFF]&memory[0xFF0F]&0x01)  // V-Blank P 1 0x0040
   { 
      memory[0xFF0F] &= 0xFE;
      PC.W = 0x0040;
      return;
   }  
  
   if((memory[0xFFFF]&memory[0xFF0F]&0x02))  // LCDC stat P 2 0x0048
   {
      memory[0xFF0F] &= 0xFD;
      PC.W = 0x0048;
      return;
   } 
               
   if(memory[0xFFFF]&memory[0xFF0F]&0x04)  // Timer Overflow P 3 0x0050
   {
      memory[0xFF0F] &= 0xFB;
      PC.W = 0x0050;
      return;
   } 
      
   if(memory[0xFFFF]&memory[0xFF0F]&0x08)  // Serial Transfer P 4 0x0058
   {
      memory[0xFF0F] &= 0xF7;
      PC.W = 0x0058;
      return;
   }
   
   if(memory[0xFFFF]&memory[0xFF0F]&0x10)  // Hi-Lo Pin(button press) P 5 0x0060
   {
      memory[0xFF0F] &= 0xEF;
      PC.W = 0x0060;
      return;
   }   
}

