/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <windows.h>
#include <stdio.h>

#include "rom.h"
#include "debug.h"
#include "types.h"
#include "config.h"
#include "main.h"
#include "mainloop.h"
#include "directdraw.h"
#include "sound.h"

#include "strings.h"
// Directories ------------------------------------------
char program_directory[ROM_PATH_SIZE];
char save_directory[SAVE_PATH_SIZE];
char state_directory[STATE_PATH_SIZE];
char rom_directory[ROM_PATH_SIZE];

extern int joy_config[6];
extern char player_name[100];

bool config_read_line(FILE* file,char* target_str)
{
   char c;
   if(target_str==NULL)
   {
      for(;;) // read comment line
      {
         c = fgetc(file);
         if(c==EOF)
            return false;
         if(c=='\n')
            return true;        
      }
   } else
   {
      int i = 0;
      for(;;)
      {
         c = fgetc(file);
         if(c==EOF)
            return false;
         if(c=='\n')
            return true;         
         target_str[i] = c;
         ++i;
      }
   } 
}

void init_options()
{
   if(sgb_border_option == SGB_BORDER)
   {
      CheckMenuItem(menu,IDM_CPUGBCSGB,MF_CHECKED);
   } else if(sgb_border_option == SGB_INITB)
   {
      CheckMenuItem(menu,IDM_CPUGBCSGBI,MF_CHECKED);
   }
   
   if(auto_frameskip)
      CheckMenuItem(menu,IDM_OPTIONVIDEOFSAUTO,MF_CHECKED);  
      
   if(gbc_real_colors)
      CheckMenuItem(menu,IDM_OPTIONVIDEOGBCCOLORS,MF_CHECKED);  
      
   switch(sound_volume)
   {
      case 2:
         CheckMenuItem(menu,IDM_OPTIONSOUNDVOL2,MF_CHECKED);       
      break;
      case 3:
         CheckMenuItem(menu,IDM_OPTIONSOUNDVOL3,MF_CHECKED);       
      break;
      case 4:
         CheckMenuItem(menu,IDM_OPTIONSOUNDVOL4,MF_CHECKED);       
      break;
      case 1:
      default:
         CheckMenuItem(menu,IDM_OPTIONSOUNDVOL1,MF_CHECKED);      
      break;
   }
   if(sound_reverse_stereo)
      CheckMenuItem(menu,IDM_OPTIONSOUNDRSTEREO,MF_CHECKED);   
   
   if(sound_lowpass)
      CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASSNONE,MF_UNCHECKED);   
   if(sound_lowpass == LOWPASS_LEVEL1)
      CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASS1,MF_CHECKED);
   else
   if(sound_lowpass == LOWPASS_LEVEL2)
      CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASS2,MF_CHECKED);

   switch(border_filter)
   {
      case FILTER_SOFT2X:
         CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSOFT2X,MF_CHECKED);  
         
         border_filter_width = 2;
         border_filter_height = 2;
      break;   
      case FILTER_SCALE2X:
         CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE2X,MF_CHECKED);  
         
         border_filter_width = 2;
         border_filter_height = 2;
      break;    
      case FILTER_SCALE3X:
         CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE3X,MF_CHECKED);  
         
         border_filter_width = 3;
         border_filter_height = 3;
      break;          
      case FILTER_BLUR:
         CheckMenuItem(menu,IDM_VIDEOFILTERBORDERBLUR,MF_CHECKED);  
         
         border_filter_width = 2;
         border_filter_height = 2;
      break;            
      case FILTER_NONE:
      default:
         border_filter = FILTER_NONE;
         
         CheckMenuItem(menu,IDM_VIDEOFILTERBORDERNONE,MF_CHECKED);
         border_filter_width = 1;
         border_filter_height = 1;         
      break;
   }
 
   switch(filter)
   {
      case FILTER_SOFT2X:
         CheckMenuItem(menu,IDM_VIDEOFILTERSOFT2X,MF_CHECKED);  
         
         filter_width = 2;
         filter_height = 2;
                   
         change_filter(); 
      break;   
      case FILTER_SCALE2X:
         CheckMenuItem(menu,IDM_VIDEOFILTERSCALE2X,MF_CHECKED);  
         
         filter_width = 2;
         filter_height = 2;
                   
         change_filter(); 
      break;    
      case FILTER_SCALE3X:
         CheckMenuItem(menu,IDM_VIDEOFILTERSCALE3X,MF_CHECKED);  
         
         filter_width = 3;
         filter_height = 3;
                   
         change_filter(); 
      break;          
      case FILTER_BLUR:
         CheckMenuItem(menu,IDM_VIDEOFILTERBLUR,MF_CHECKED);  
         
         filter_width = 2;
         filter_height = 2;
                   
         change_filter(); 
      break;            
      case FILTER_NONE:
      default:
         filter = FILTER_NONE;
         
         CheckMenuItem(menu,IDM_VIDEOFILTERNONE,MF_CHECKED);
         filter_width = 1;
         filter_height = 1;        
         
          change_filter(); 
      break;
   }
   
   if(use_mmx)
      CheckMenuItem(menu,IDM_VIDEOFILTERUSEMMX,MF_CHECKED);
   
   if(!unk_opcode_option)
      CheckMenuItem(menu,IDM_CPUOPCODE,MF_UNCHECKED);
   
   if(zip_name_option)
      CheckMenuItem(menu,IDM_FILEZIPNAME,MF_CHECKED);
   
   switch(use_joystick_input)
   {
     case 0: CheckMenuItem(menu,IDM_OPTIONCONTROLJOY1,MF_CHECKED); break;
     case 1: CheckMenuItem(menu,IDM_OPTIONCONTROLJOY2,MF_CHECKED); break;
     case 2: CheckMenuItem(menu,IDM_OPTIONCONTROLJOY3,MF_CHECKED); break;
     case 3: CheckMenuItem(menu,IDM_OPTIONCONTROLJOY4,MF_CHECKED); break;
   }
   
   CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
   CheckMenuItem(menu,IDM_OPTIONVIDEOFS0+frame_skip,MF_CHECKED); 
   
   if(!sprite_limit)
      CheckMenuItem(menu,IDM_OPTIONVIDEOSPRLIM,MF_UNCHECKED);   
      
   if(!LCDoff_option)
      CheckMenuItem(menu,IDM_OPTIONVIDEOLCDOFF,MF_UNCHECKED);
      
   if(speed_up_option&SP_FS9)
      CheckMenuItem(menu,IDM_CPUSPFS9,MF_CHECKED);

   if(speed_up_option&SP_SNDOFF)
      CheckMenuItem(menu,IDM_CPUSPSNDOFF,MF_CHECKED);
   
   if(!system_sound_on)
      CheckMenuItem(menu,IDM_OPTIONSOUND,MF_UNCHECKED);   
    
   switch(color_option)
   {
      case COL_LCDBROWN:
         CheckMenuItem(menu,IDM_OPTIONVIDEOLCDBROWN,MF_CHECKED);  
      break; 
      case COL_LCDGREEN:
         CheckMenuItem(menu,IDM_OPTIONVIDEOLCDGREEN,MF_CHECKED);  
      break;
      case COL_GRAY:
         CheckMenuItem(menu,IDM_OPTIONVIDEOGRAY,MF_CHECKED);  
      break;            
      case COL_BW:
      default:
         CheckMenuItem(menu,IDM_OPTIONVIDEOBW,MF_CHECKED);
      break;	 
   }
   if(mix_frames)
   {
      CheckMenuItem(menu,IDM_OPTIONVIDEOMIX,MF_CHECKED);

      if(dx_bitcount == 16)
         draw_screen = draw_screen_mix16;
      else
         draw_screen = draw_screen_mix32;             	              
   } else
   {
      CheckMenuItem(menu,IDM_OPTIONVIDEOMIX,MF_UNCHECKED);

      if(dx_bitcount == 16)
         draw_screen = draw_screen16;
      else
         draw_screen = draw_screen32;             	                           
   }
   
   RECT winRect;
   GetWindowRect(hwnd,&winRect);
   if(win_size <= 0 || win_size > 4)
      win_size=2;
   MoveWindow(hwnd,winRect.left,winRect.top,win_size*160+sizen_w,win_size*144+sizen_h,TRUE);
}

bool read_config_file()
{
   SetCurrentDirectory(program_directory);
   
   strcpy(save_directory,program_directory);
   strcat(save_directory,"\\save");

   strcpy(state_directory,program_directory);
   strcat(state_directory,"\\state");
       
   strcpy(rom_directory,program_directory);
   
   FILE* configfile = fopen("gest.cfg","rb");
   if(!configfile) 
   { 
      RECT winRect;
      GetWindowRect(hwnd,&winRect);

      win_size = 2;
      MoveWindow(hwnd,winRect.left,winRect.top,win_size*160+sizen_w,win_size*144+sizen_h,TRUE);

      init_options();
      return true; 
   }
   memset(rom_directory,0,ROM_PATH_SIZE);
   memset(save_directory,0,SAVE_PATH_SIZE);
   memset(state_directory,0,STATE_PATH_SIZE);
   
   if(!config_read_line(configfile,NULL)) return false;
   if(!config_read_line(configfile,rom_directory)) return false;
   
   if(!config_read_line(configfile,NULL)) return false;
   if(!config_read_line(configfile,save_directory)) return false;   

   if(!config_read_line(configfile,NULL)) return false;
   if(!config_read_line(configfile,state_directory)) return false;

   fread(&color_option, sizeof(int),1,configfile); 
   fread(&win_size, sizeof(int),1,configfile);
   fread(&mix_frames, sizeof(int),1,configfile); 
   fread(multi_key_config, sizeof(int),4*17,configfile);
   fread(&speed_up_option,sizeof(int),1,configfile);
   fread(&frame_skip,sizeof(int),1,configfile);
   fread(&sprite_limit,sizeof(bool),1,configfile);
   fread(&LCDoff_option,sizeof(int),1,configfile);
   fread(&system_sound_on,sizeof(int),1,configfile);
   fread(&use_joystick_input,sizeof(int),1,configfile);
   fread(&zip_name_option,sizeof(int),1,configfile);
   fread(&unk_opcode_option,sizeof(int),1,configfile);
   fread(&filter,sizeof(int),1,configfile);
   fread(&use_mmx,sizeof(int),1,configfile);
   fread(&sound_lowpass,sizeof(int),1,configfile);
   fread(&sound_reverse_stereo,sizeof(int),1,configfile);     
   fread(&sound_volume,sizeof(int),1,configfile);     
   fread(&border_filter,sizeof(int),1,configfile);
   fread(&gbc_real_colors,sizeof(int),1,configfile);
   fread(joy_config, sizeof(int),6,configfile);
   fread(&auto_frameskip, sizeof(int),1,configfile);
   fread(&sgb_border_option, sizeof(int),1,configfile);
   
   init_options();
   
   fclose(configfile);
   return true;
}

bool write_config_file()
{
   SetCurrentDirectory(program_directory);
   
   FILE* configfile = fopen("gest.cfg","wb");
   if(!configfile) 
   { 
      debug_print(str_table[ERROR_CFG_FILE]);
      return false; 
   }
       
   fputs("#Rom directory:\n",configfile);
   fputs(rom_directory,configfile);
   fputc('\n',configfile);
   fputs("#SRAM save directory:\n",configfile);
   fputs(save_directory,configfile);
   fputc('\n',configfile);
   fputs("#Save State directory:\n",configfile);
   fputs(state_directory,configfile);
   fputc('\n',configfile);
   fwrite(&color_option, sizeof(int),1,configfile); 
   fwrite(&win_size, sizeof(int),1,configfile);
   fwrite(&mix_frames, sizeof(int),1,configfile); 
   fwrite(multi_key_config, sizeof(int),4*17,configfile);
   fwrite(&speed_up_option,sizeof(int),1,configfile);
   fwrite(&frame_skip,sizeof(int),1,configfile);
   fwrite(&sprite_limit,sizeof(bool),1,configfile);
   fwrite(&LCDoff_option,sizeof(int),1,configfile);
   fwrite(&system_sound_on,sizeof(int),1,configfile);
   fwrite(&use_joystick_input,sizeof(int),1,configfile);
   fwrite(&zip_name_option,sizeof(int),1,configfile);
   fwrite(&unk_opcode_option,sizeof(int),1,configfile);
   fwrite(&filter,sizeof(int),1,configfile);
   fwrite(&use_mmx,sizeof(int),1,configfile);
   fwrite(&sound_lowpass,sizeof(int),1,configfile);
   fwrite(&sound_reverse_stereo,sizeof(int),1,configfile);     
   fwrite(&sound_volume,sizeof(int),1,configfile);     
   fwrite(&border_filter,sizeof(int),1,configfile);
   fwrite(&gbc_real_colors,sizeof(int),1,configfile);
   fwrite(joy_config, sizeof(int),6,configfile);
   fwrite(&auto_frameskip, sizeof(int),1,configfile);
   fwrite(&sgb_border_option, sizeof(int),1,configfile);
         
   fclose(configfile);
   return true;
}

