/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
    VisualBoyAdvance - a Gameboy and GameboyAdvance emulator
    Copyright (C) 1999-2004 by Forgotten

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
 
#include "ioregs.h"
#include <stdio.h>
#include <stdlib.h>

#include <dplay.h>
#include "directplay.h"
#include "devices.h"
#include "cpu.h"
#include "mem.h"
#include "SGB.h"
#include "directinput.h"
#include "sound.h"
#include "MBC.h"
#include "GB.h"
#include "mainloop.h"

//extern int skip_frame;
// I/O Regs read common for all readmemory functions
byte io_reg_read(register unsigned short address)
{         
   if(address >= 0xFF00)
   {
      switch(address&0x00FF)
      {
      case 0x00:
      {
         if(sgb_mode) 
         {
            sgb_readingcontrol |= 4;
            sgb_reset_state();
                     
            if(sgb_multiplayer)
            {
               switch(sgb_nextcontrol)
               {
               case 0x0F:
                  Check_KBInput(0);
               break;
               case 0x0E:
                  Check_KBInput(1);
               break;
               case 0x0D:
                  Check_KBInput(2);
               break;
               case 0x0C:
                  Check_KBInput(3);
               break;
               }
            }
         } 
         
         int* button_pressed = GB->button_pressed;
         if((GB->memory[0xFF00]&0x30) == 0x00) //P14 and P15 not set
            GB->memory[0xFF00] = (0xC0|(button_pressed[B_A]|button_pressed[B_RIGHT])|((button_pressed[B_B]|button_pressed[B_LEFT])<<1)|((button_pressed[B_SELECT]|button_pressed[B_UP])<<2)|((button_pressed[B_START]|button_pressed[B_DOWN])<<3));         
         else 
         if((GB->memory[0xFF00]&0x30) == 0x20) //P14 not set
            GB->memory[0xFF00] = (0xE0|button_pressed[B_RIGHT]|(button_pressed[B_LEFT]<<1)|(button_pressed[B_UP]<<2)|(button_pressed[B_DOWN]<<3));
         else 
         if((GB->memory[0xFF00]&0x30) == 0x10) //P15 not set
            GB->memory[0xFF00] = (0xD0|button_pressed[B_A]|(button_pressed[B_B]<<1)|(button_pressed[B_SELECT]<<2)|(button_pressed[B_START]<<3));
         else
            if(sgb_mode && sgb_multiplayer) 
               GB->memory[0xFF00] = 0xF0 | sgb_nextcontrol;
            else 
               GB->memory[0xFF00] = 0xFF;

         return GB->memory[0xFF00];
      }
      break;
      
      case 0x30: case 0x31: case 0x32: case 0x33: case 0x34: case 0x35: case 0x36:
      case 0x37: case 0x38: case 0x39: case 0x3A: case 0x3B: case 0x3C: case 0x3D: 
      case 0x3E: case 0x3F: // waveform data
         if(channel3_on)
            return 0xFF;
         else
            return GB->memory[address];
      break;

      case 0x56: // RP infrared
      if(GB->gbc_mode)
      {  
         if((GB->memory[0xFF56]&0xC0) == 0xC0) // reading data
            return (GB->memory[0xFF56]&0xC1)|0;
         else
            return GB->memory[0xFF56];
      }
      break;
      // undocumented GBC registers
      case 0x6C:
         if(GB->gbc_mode)
            return (GB->memory[0xFF6C]&0x01);
      break;
      case 0x75:
         if(GB->gbc_mode)
            return (GB->memory[0xFF75]&0x70);
      break;     
      }
   }
      
   return GB->mem_map[address>>12][address&0x0FFF];
}

extern int time_to_switch;
extern int data_trans;
// I/O regs and memory echo common for all writememory functions
int io_reg_write(register unsigned short address,register byte data)
{
 /*  if(address >= 0x8000 && address < 0xA000)
   {
      if(LCDon && (memory[0xFF41]&0x03) == 0x03)
         return 1;
   }  
   if(address >= 0xFE00 && address < 0xFEA0)
   {
      if(LCDon && (memory[0xFF41]&0x02))
         return 1;   
   }
   if(address >= 0xFEA0 && address < 0xFF00)
      return 1;
   if(gbc_mode && LCDon && (address==0xFF69 || address==0xFF6B) && (memory[0xFF41]&0x03)==3)
     return 1; */
      
   // memory echo
   if(address >= 0xE000 && address < 0xFE00)
      GB->mem_map[(address-0x2000)>>12][(address-0x2000)&0x0FFF] = data;
   if(address >= 0xC000 && address < 0xDE00)
      GB->memory[(address+0x2000)] = data;

   if(address >= 0xFF00)
   {
      switch(address&0x00FF)
      {
      case 0x00: // P1
         if(sgb_mode && sgb_file_transfer)
            sgb_transfer(data);
            
         GB->memory[0xFF00] = ((GB->memory[0xFF00] & 0xCF)|(data&0x30));
      return 1;
      case 0x01: // SB
         GB->memory[0xFF01] = data;
      return 1;
      case 0x02: // SC serial control
      {
         int old_sc = GB->memory[0xFF02];
         GB->memory[0xFF02] = data;  
         time_to_switch = 1;
  
         if(data&0x80)
         {              
            if(!GB->gbc_mode || !(data&2))
               GB->CYCLES_SERIAL = CYCLES_SERIAL_GB;
            else if(GB->gbc_mode && (data&2))
               GB->CYCLES_SERIAL = CYCLES_SERIAL_GBC;

            if(net_initted)
            {
               if(data&1)
               {
                  if(!link_on) check_link_state(1,1);

                  if(link_on)
                  {
                     send_gb_link_state(1,1);
                  }

                  data_sent = 0;
                  GB->serialbits = 0;
                  GB->cycles_serial = GB->CYCLES_SERIAL;
               } else
               {
                  data_sent = 0;
                  link_on = 0; // important!!
               
                  //if(!(old_sc&0x80) || (old_sc&1)) 
                  {
                     send_gb_link_state(1,0);
                     //link_send_data(memory[0xFF01]);
                  }

                  GB->serialbits = 7;
                  GB->cycles_serial = 20;//CYCLES_SERIAL;
               }
            } else
            {
               GB->serialbits = 0;
               GB->cycles_serial = 8*GB->CYCLES_SERIAL;
           /*    if(multiple_gb && ((GB->gb_speed&&gb_speed_another) || ((data&1) && !(memory_another[0xFF02]&0x80))))// && !(memory_another[0xFF02]&1))
               {
                  time_to_switch = 1;                                
               } else
               if(multiple_gb && (data&0x80) && !(data&1))// && (memory_another[0xFF02]&0x81))
               {
                  time_to_switch = 1;
               }*/
          /*     else
               if(multiple_gb && !(data&1) && (memory_another[0xFF02]&0x80))
               {
                  time_to_switch = 1;
               }*/
            }
         }
      }
      return 1;
      case 0x04: // DIV  
         GB->memory[0xFF04] = 0x00;
      return 1;  
      case 0x05: // TIMA
         GB->memory[0xFF05] = data;        
      return 1;
      case 0x06: // TMA
         GB->memory[0xFF06] = data;    
      return 1;      
      case 0x07: // TAC    
         GB->memory[0xFF07] = 0xF8|data;
         GB->timeron = (data&0x04);
         switch(data&0x03)
         {
            case 0: 
               GB->timer_freq = CYCLES_TIMER_MODE0;
               if(GB->memory[0xFF06] != 0xd8) // fix for Korodice
                  GB->cycles_timer = GB->timer_freq; 
            break;
            case 1: GB->cycles_timer = GB->timer_freq = CYCLES_TIMER_MODE1; break;
            case 2: GB->cycles_timer = GB->timer_freq = CYCLES_TIMER_MODE2; break;
            case 3: GB->cycles_timer = GB->timer_freq = CYCLES_TIMER_MODE3; break;
         }        
      return 1;
      case 0x0F: // IF
         GB->memory[0xFF0F] = 0xE0|(data&0x1F);
      return 1;
      
      case 0x10:
      case 0x11:
      case 0x12:
      case 0x13:
      case 0x14:
      case 0x16:
      case 0x17:
      case 0x18:
      case 0x19:
      case 0x1a:
      case 0x1b:
      case 0x1c:
      case 0x1d:
      case 0x1e:
      case 0x20:
      case 0x21:
      case 0x22:
      case 0x23:
      case 0x24:
      case 0x25:
      case 0x26: 
      {
         sound_register(address,data);
         return 1;
      }
      
      case 0x30: case 0x31: case 0x32: case 0x33: case 0x34: case 0x35: case 0x36:
      case 0x37: case 0x38: case 0x39: case 0x3A: case 0x3B: case 0x3C: case 0x3D: 
      case 0x3E: case 0x3F: // waveform data
            GB->memory[address] = data; 
      return 1;
      
      case 0x40: // LCDC        
         if((GB->memory[0xFF40]&0x80) ^ (data&0x80)) // LCD display is turned on/off
         {                          
            if(data&0x80) // on
            {
               GB->LCDon = 1;      
               GB->cycles_LCD = CYCLES_LCD_MODE2+268;
               GB->skip_frame = (GB->gbc_mode?1:4);  
               // original GB has slower LCD, so don't draw next 4 frames to
               // prevent wrong graphics display                                           

               GB->set_LCD_mode(2);
               GB->memory[0xFF44] = 0;    
            } else // off
            {           
               GB->LCDon = 0;

               GB->off_counter = 456;
               frames = 1;
                           
               GB->memory[0xFF41] &= 0xFC;                   
               GB->memory[0xFF44] = 0;
            }
         }  
              
         if(!(GB->memory[0xFF40]&0x20) && (data&0x20) && GB->windowline == -1 && (GB->memory[0xFF44] > GB->memory[0xFF4A] && GB->memory[0xFF44] < 144))     
            GB->windowline = 154; // don't draw window                

         GB->memory[0xFF40] = data;

         GB->tile_map = ((GB->memory[0xFF40] & 0x08)?0x1c00:0x1800);
         GB->win_tile_map = ((GB->memory[0xFF40] & 0x40)?0x1c00:0x1800);
         GB->tile_pattern = ((GB->memory[0xFF40] & 0x10)?0x0000:0x0800);  
      return 1;
      case 0x41: // STAT
         GB->memory[0xFF41] = 0x80|((GB->memory[0xFF41]&0x07)|(data&0x78));
         
         GB->int_line = -1; //??? fix for Ken Griffey Jr's Slugfest
         
         if(!GB->gbc_mode && GB->LCDon && (GB->memory[0xFF41]&0x03)==1 && GB->memory[0xFF44] != 0) // DMG bug        
            GB->set_int(2);  
      return 1;
      case 0x42: // SCY
         GB->memory[0xFF42] = data;         
      return 1;       
      case 0x43: // SCX
         GB->memory[0xFF43] = data;
      return 1;             
      case 0x44: // LY   
          //It's read only
      return 1; 
      case 0x45: // LYC
      {                 
         GB->memory[0xFF45] = data;

         if(GB->LCDon) // can cause LYC interrupt
         { 
            GB->compareLYCtoLY();
            if((GB->memory[0xFF41] & 0x40) && (GB->memory[0xFF41] & 0x04))
               GB->set_int(2);
         }
      }
      return 1;
      case 0x46: // DMA transfer 
         copy_memory(0xFE00,data<<8,0xA0);
         GB->memory[0xFF46] = data;
      return 1; 
      case 0x47: // BG & W palette
         GB->BGP[0] = (data & 0x03);
         GB->BGP[1] = (data & 0x0c)>>2;
         GB->BGP[2] = (data & 0x30)>>4;
         GB->BGP[3] = (data & 0xc0)>>6; 
         GB->memory[0xFF47] = data; 
      return 1;
      case 0x48: // OBJ palette 0
         //GB->OBP0[0] = (data & 0x03); //it's transparent
         GB->OBP0[1] = (data & 0x0c)>>2;
         GB->OBP0[2] = (data & 0x30)>>4;
         GB->OBP0[3] = (data & 0xc0)>>6;
         GB->memory[0xFF48] = data; 
      return 1;
      case 0x49: // OBJ palette 1
         //GB->OBP1[0] = (data & 0x03); //it's transparent
         GB->OBP1[1] = (data & 0x0c)>>2;
         GB->OBP1[2] = (data & 0x30)>>4;
         GB->OBP1[3] = (data & 0xc0)>>6; 
         GB->memory[0xFF49] = data; 
      return 1;
      case 0x4A: // WY                               //?
         if((GB->memory[0xFF40]&0x20) && GB->memory[0xFF44] < 10 && GB->memory[0xFF44] > GB->memory[0xFF4A])// && data > 80 && data!=128)
         {
            GB->wy_set = data;
            return 1;
         }
            //fix for Razor
         if(GB->memory[0xFF44] != 89 && GB->memory[0xFF44] < 144 && GB->windowline == -1 && data < GB->memory[0xFF44] && (GB->memory[0xFF40]&0x20)) 
            GB->windowline = 154; // fix for Puyo Wars,Urusei Yatsura      
         else if(GB->windowline == 154 && data > GB->memory[0xFF44]) 
            GB->windowline = 0; // Zen Intergalactic Ninja 
 
         GB->memory[0xFF4A] = data; 
      return 1;             
      case 0x4B: // WX          
         GB->memory[0xFF4B] = data;                         
      return 1;             
      case 0x4D: // KEY1 speed switch
      if(GB->gbc_mode)
      {
         GB->memory[0xFF4D] = (GB->memory[0xFF4D] & 0x80) | (data & 1);
         return 1;
      }
      break;
      case 0x4F: // VBK //switch video RAM bank
      if(GB->gbc_mode)
      {
         data = (data&1);
         GB->vram_bank = data;
         int vramadr = data*0x2000;
         GB->mem_map[0x08] = &GB->VRAM[vramadr];
         GB->mem_map[0x09] = &GB->VRAM[vramadr + 0x1000];
         GB->memory[0xFF4F] = 0xFE|data;
         return 1;
      }
      break;
      case 0x51: // HDMA1
      if(GB->gbc_mode)
      {      
         if(data > 0x7f && data < 0xa0)
            data = 0;

         GB->hdma_source = (data << 8) | (GB->memory[0xFF52] & 0xf0);
      
         GB->memory[0xFF51] = data;
         return 1;
      }
      break;
      case 0x52: // HDMA2
      if(GB->gbc_mode)
      {        
         data = data & 0xf0;
      
         GB->hdma_source = (GB->memory[0xFF51] << 8) | (data);
      
         GB->memory[0xFF52] = data;
         return 1;
      }
      break;
      case 0x53: // HDMA3
      if(GB->gbc_mode)
      {        
         data = data & 0x1f;
         GB->hdma_destination = (data << 8) | (GB->memory[0xFF54] & 0xf0);
         GB->hdma_destination += 0x8000;
         
         GB->memory[0xFF53] = data;
         return 1;
      }
      break;
      case 0x54: // HDMA4
      if(GB->gbc_mode)
      {
         data = data & 0xf0;
         GB->hdma_destination = ((GB->memory[0xFF53] & 0x1f) << 8) | data;
         GB->hdma_destination += 0x8000;
         
         GB->memory[0xFF54] = data;
         return 1;
      }
      break;
      case 0x55: // HDMA5
      if(GB->gbc_mode)
      {
         GB->hdma_bytes = ( (data & 0x7F) + 1) << 4;
         if(GB->hdma_on) 
         {
            if(data & 0x80) 
            {
               GB->memory[0xFF55] = (data & 0x7F);
            } else 
            {
               GB->memory[0xFF55] = 0xFF;
               GB->hdma_on = 0;
            }
         } else 
         {
            if(data & 0x80) // HDMA
            {
               GB->hdma_on = 1;
               GB->memory[0xFF55] = data & 0x7F;
               if((GB->memory[0xFF41]&0x03) == 0)
                  GB->do_hdma();
            } else // GDMA
            {  
               GB->gdma_rest = (((data&0x7F)+1)*8); // is this OK?
                                             
               copy_memory(GB->hdma_destination,GB->hdma_source,GB->hdma_bytes);

               GB->hdma_destination += GB->hdma_bytes;
               GB->hdma_source += GB->hdma_bytes;
          
               GB->memory[0xFF53] = ((GB->hdma_destination - 0x8000) >> 8) & 0x1F;
               GB->memory[0xFF54] = GB->hdma_destination & 0xF0;
               GB->memory[0xFF51] = (GB->hdma_source >> 8) & 0xFF;
               GB->memory[0xFF52] = GB->hdma_source & 0xF0;
            }
         }
         return 1;
      }
      break;
      case 0x56: // RP infrared
      if(GB->gbc_mode)
      {
         GB->memory[0xFF56] = (GB->memory[0xFF56]&0x02)|(data&0xFD);
         return 1;
      }
      break;
      case 0x68: // BCPS
      if(GB->gbc_mode)
      {
         if(!GB->rom->CGB) // this is a GBC game
            GB->rom->CGB = 1;
            
         int index = (data & 0x3f) >> 1;
         int hilo  = (data & 0x01);
      
         GB->memory[0xff68] = data;
         GB->memory[0xff69] = (hilo ? (GB->GBC_BGP[index] >> 8) : (GB->GBC_BGP[index] & 0x00ff));
         return 1;
      }
      break;
      case 0x69: // BCPD
      if(GB->gbc_mode)
      {
         int index = (GB->memory[0xFF68] & 0x3f) >> 1;
         int hilo = (GB->memory[0xFF68] & 1);
         
         GB->memory[0xFF69] = data;
         GB->GBC_BGP[index] = (hilo ? ((data << 8) | (GB->GBC_BGP[index] & 0xff)) : ((GB->GBC_BGP[index] & 0xff00) | data)) & 0x7fff;
         
         if(GB->memory[0xff68] & 0x80) 
         {
            int pindex = ((GB->memory[0xff68] & 0x3f) + 1) & 0x3f;
        
            GB->memory[0xff68] = (GB->memory[0xff68] & 0x80) | pindex;
            GB->memory[0xff69] = (pindex & 1 ? (GB->GBC_BGP[pindex>>1] >> 8) : (GB->GBC_BGP[pindex>>1] & 0x00ff));
         }
         return 1;
      }
      break;
      case 0x6A: // OCPS
      if(GB->gbc_mode)
      {    
         int index = (data & 0x3f) >> 1;
         int hilo  = (data & 0x01);
      
         GB->memory[0xff6A] = data;
         GB->memory[0xff6B] = (hilo ? (GB->GBC_OBP[index] >> 8) : (GB->GBC_OBP[index] & 0x00ff));
         return 1;
      }
      break;
      case 0x6B: // OCPD
      if(GB->gbc_mode)
      {   
         int index = (GB->memory[0xFF6A] & 0x3f) >> 1;
         int hilo = (GB->memory[0xFF6A] & 1);
         
         GB->memory[0xFF6B] = data;
         GB->GBC_OBP[index] = (hilo ? ((data << 8) | (GB->GBC_OBP[index] & 0xff)) : ((GB->GBC_OBP[index] & 0xff00) | data)) & 0x7fff;
         
         if(GB->memory[0xFF6A] & 0x80) 
         {
            int pindex = ((GB->memory[0xFF6A] & 0x3f) + 1) & 0x3f;
        
            GB->memory[0xFF6A] = (GB->memory[0xFF6A] & 0x80) | pindex;
            GB->memory[0xFF6B] = (pindex & 1 ? (GB->GBC_OBP[pindex>>1] >> 8) : (GB->GBC_OBP[pindex>>1] & 0x00ff));
         }
         return 1;
      }
      break;
      case 0x70: // SVBK // switch internal RAM bank
      if(GB->gbc_mode)
      {
         data = (data&7);
         int bank = data;
         if(data == 0)
            bank = 1;
         
         GB->wram_bank = bank;
         
         GB->mem_map[0x0D] = &GB->WRAM[bank*0x1000];

         GB->memory[0xFF70] = 0xF8|data;
         return 1;
      }
      break;
      
      // undocumented GBC registers
      case 0x6C: 
      if(GB->gbc_mode)
      {
         GB->memory[0xFF6C] = data&0x01;
         return 1;
      }
      break;
      case 0x72: 
      case 0x73:
      case 0x74:
      if(GB->gbc_mode)
      {
         GB->memory[address] = data;
         return 1;
      }
      break;               
      case 0x75: 
      if(GB->gbc_mode)
      {
         GB->memory[0xFF75] = data&0x70;
         return 1;
      }
      break;
      case 0x77: 
      case 0x76: 
      if(GB->gbc_mode)
      {
         // read only
         return 1;
      }
      break;
      case 0xFF: // IE
         GB->memory[0xFFFF] = data&0x1F;
      return 1;
      }
   }
   return 0;
}

