/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
    VisualBoyAdvance - a Gameboy and GameboyAdvance emulator
    Copyright (C) 1999-2004 by Forgotten

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <string.h>
#include <stdio.h>

#include "mainloop.h"
#include "mem.h"
#include "MBC.h"
#include "render.h"
//#include <dplay.h>
//#include "directplay.h"
#include "types.h"
#include "debug.h"
#include "cpu.h"
#include "rom.h"
#include "gfx.h"
#include "SGB.h"
#include "sound.h"
#include "devices.h"
#include "GB.h"

#include "strings.h"

bool sprite_limit = true;
int LCDoff_option = 1;
int gbc_real_colors = 0;

int unk_opcode_option = 1;
//int button_pressed[8];
int last_button_pressed[8];
int sensor_dir[4];
int sensorX = 2047;
int sensorY = 2047;

int frames = 0;
int frame_skip = 0;

int real_cycle = 0; // for gbc 2xspeed mode

void reset_no_change() // don't change modes
{
   if(sgb_mode)
   {
      fill_gfx_buffers(0UL);             
   } else   
   if(GB->gbc_mode)
   {
      if(GB->gfx_bit_count==16)
         fill_gfx_buffers(gfx_pal16[0x7fff]|(gfx_pal16[0x7fff]<<16)); 
      else
         fill_gfx_buffers(gfx_pal32[0x7fff]);             
   } else   
   {
      if(GB->gfx_bit_count==16)
         fill_gfx_buffers(gfx_pal16[GB->GBC_BGP[0]]|(gfx_pal16[GB->GBC_BGP[0]]<<16)); 
      else
         fill_gfx_buffers(gfx_pal32[GB->GBC_BGP[0]]);              
   }
   
   if(gbc_real_colors && GB->gbc_mode)
      mix_gbc_colors();
         
   if(GB->gbc_mode)
      draw_line_tile = draw_line_tile_GBC;
   else
      draw_line_tile = draw_line_tile_DMG;
      
   memset(GB->gfx_buffer,0x00,160*144*2);
   memset(GB->gfx_buffer_old,0x00,160*144*2);
    
   sensor_dir[0] = sensor_dir[1] = sensor_dir[2] = sensor_dir[3] = 0;

   frames = 0;

   GB1->reset();
   
   reset_devices();
       
   cpu_reset(1);
   
   mem_reset(1);
   
   sound_reset();
}

void reset(int gb_number)
{     
   if(!multiple_gb && gb_number==2) return;
   int switched = 0;
   if(GB1->romloaded && GB2 && GB2->romloaded && current_gb!=gb_number)
   {
      switch_gb();
      switched = 1;
   }

   border_uploaded = 0;
         
   sensor_dir[0] = sensor_dir[1] = sensor_dir[2] = sensor_dir[3] = 0;
   
   frames = 0;
      
   reset_devices();

   cpu_reset(gb_number);
   
   mem_reset(gb_number);
   
   if(gb_number==1)
      sound_reset();
   
   if(gb_number == 1)
      GB1->reset();
   else
      GB2->reset();
   
   if(switched)
      switch_gb();
}


////////////////////////////////////////
//int mainloop()
//-the emulator main loop
////////////////////////////////////////
int data_trans = 0;
byte data_sended;
int time_to_switch = 0;

int gb_system::mainloop()
{
while(frames==0 && time_to_switch == 0)
{
   if(multiple_gb && data_trans && (memory[0xFF02]&0x80) && !(memory[0xFF02]&1))
   {
      memory[0xFF01] = data_sended;
      cycles_serial = 0;
      memory[0xFF02] &= 0x7F;
      set_int(8);
      serialbits = 0;
      data_trans = 0;
      time_to_switch = 1;   
   }

   if(CPUHalt == HALT || gdma_rest)
   { 
      cur_cycle = (4>>gb_speed);
      real_cycle = 4;
      if(gdma_rest)
      {
         gdma_rest -= cur_cycle;
         if(gdma_rest < 0)
            gdma_rest = 0;
      }
   } else 
   { 
      opcode = readopcode(PC.W++);

      cur_cycle = cycles[opcode];    

      if(CPUHalt == HALT2)
      {
         CPUHalt = 0;
         --PC.W;
      }
      
      #include "cpucore.h"

      real_cycle = cur_cycle;
      cur_cycle >>= gb_speed;
   }

   if(sgb_mode && cycles_SGB) // SGB command timeout
   {
      cycles_SGB -= cur_cycle;
      if(cycles_SGB <= 0)
      {
         cycles_SGB = 0;
         sgb_reset_state();

         cycles_SGB += CYCLES_SGB_TIMEOUT;
      }
   }
   
   cycles_DIV -= real_cycle;
   if(cycles_DIV <= 0)
   {         
      ++memory[0xFF04];
      cycles_DIV += CYCLES_DIV; 
   }               
         
   if(timeron) // Timer emulation
   {
      cycles_timer -= real_cycle;
      if(cycles_timer <= 0)
      {
         ++memory[0xFF05];
         if(memory[0xFF05] == 0x00)
         {
            memory[0xFF05] = memory[0xFF06]; // Load timer modulo
            set_int(4); // Timer interrupt
         }
         cycles_timer += timer_freq; 
      }
   }

   if(LCDon) // LCD emulation
   {
      cycles_LCD -= cur_cycle;

      if(cycles_LCD <= 0)
      {         
         switch(memory[0xFF41]&0x03) // Which mode has ENDED?
         { 
         case 0: // H-Blank         
            ++memory[0xFF44];
            compareLYCtoLY();
            if(multiple_gb && !(memory[0xFF44]%5)) time_to_switch = 1;

            if(memory[0xFF44] >= 144) // time for V-Blank?
            {                                                
               set_LCD_mode(1); // next: MODE 1
                                                  
               int r_frame_skip = (speedup&&(speed_up_option&SP_FS9)?9:frame_skip);
               if(!skip_frame && !sgb_mask && !(frameskip_counter%(r_frame_skip+1)))              
                  draw_screen();
                  
               if(wy_set != -1)
               {
                  memory[0xFF4A] = wy_set;
                  wy_set = -1;
               }
                  
               if(skip_frame)
                  --skip_frame;
   
               ++frameskip_counter;
               
               ++frames;                                        

               windowline = -1;               
               
               IWait = 1;          

               if(gbc_mode)
                  cycles_LCD += 22;
               else
                  cycles_LCD += 28;      
               break;
            }  
                          
            set_LCD_mode(2); // next: MODE 2
            cycles_LCD += CYCLES_LCD_MODE2;

            if((memory[0xFF41] & 0x40) && (memory[0xFF41] & 0x04))
               STAT_wait = 36; 
   
            if(!(memory[0xFF41] & 0x40) && (memory[0xFF41]&0x20))
               set_int(2);                                 
         break;
         case 1: // V-Blank                                                            			
            if(IWait == 1)
            {
               int_line = -1;
               if(gbc_mode)
		            cycles_LCD += 434;	  
               else
                  cycles_LCD += 428;
                           
               set_int(1);
                              
               if(memory[0xFF41]&0x10)
                  set_int(2);
        
               if((memory[0xFF41] & 0x40) && (memory[0xFF41] & 0x04))
                  set_int(2);
               
               if(!sgb_multiplayer)
                  Check_KBInput(0);
                  
                // joypad interrupt
               if((memory[0xFF00]&0x10) || ((memory[0xFF00]&0x30) == 0x00))
               {
                  if(!button_pressed[B_SELECT] || !button_pressed[B_START] || !button_pressed[B_B] || !button_pressed[B_A])
                     set_int(0x10); 
                  else 
                  if(!last_button_pressed[B_SELECT] || !last_button_pressed[B_START] || !last_button_pressed[B_B] || !last_button_pressed[B_A])
                     set_int(0x10);                            
               }
               if((memory[0xFF00]&0x20) || ((memory[0xFF00]&0x30) == 0x00))
               {
                  if(!button_pressed[B_UP] || !button_pressed[B_DOWN] || !button_pressed[B_LEFT] || !button_pressed[B_RIGHT])
                     set_int(0x10);               
                  else 
                  if(!last_button_pressed[B_UP] || !last_button_pressed[B_DOWN] || !last_button_pressed[B_LEFT] || !last_button_pressed[B_RIGHT])
                     set_int(0x10);                                    
               }
               memcpy(last_button_pressed,button_pressed,8*sizeof(int));    
        
               IWait = 0;
			   
               if(rom->bankType == MBC7)
                  check_sensor();
			   
               break;
            }

            if(memory[0xFF44] == 0)
            {
               //compareLYCtoLY();

               set_LCD_mode(2); // next: MODE 2
               cycles_LCD += CYCLES_LCD_MODE2;
 
               //if(memory[0xFF41]&0x20)
               //   set_int(2);

               //if((memory[0xFF41] & 0x40) && (memory[0xFF41] & 0x04))
               //   set_int(2);
               break;
            } 
            
            if(memory[0xFF44] < 153)
            {
               ++memory[0xFF44];
               if(memory[0xFF44] >= 153)
                  cycles_LCD += 24;
               else
                  cycles_LCD += CYCLES_LCD_MODE1;
            } else
            { 
               memory[0xFF44] = 0;         
               cycles_LCD += 432+18;
            }
            
            if(gbc_mode && memory[0xFF44] >= 153 && !(memory[0xFF41] & 0x40) && (memory[0xFF41]&0x08))
               STAT_wait = CYCLES_LCD_MODE0-58;

            if(multiple_gb && !(memory[0xFF44]%40)) time_to_switch = 1;

            compareLYCtoLY();

            if((memory[0xFF41] & 0x04) && (memory[0xFF41] & 0x40))
               set_int(2);
         break;
         case 2: // searching OAM-RAM       
            draw_wait = 65; //48 //54  

            if(memory[0xFF45] == memory[0xFF44] && memory[0xFF4A] > 130 && memory[0xFF4A] < 144)
               draw_wait = 180-cycles_LCD; // ???? // fix for Parodius

            set_LCD_mode(3); // next: MODE 3
            mode3_cyc = (CYCLES_LCD_MODE3);//+12*sprite_number);
            cycles_LCD += mode3_cyc;               
         break;
         case 3: // transferring data to LCD driver      
            set_LCD_mode(0); // next: MODE0 
            cycles_LCD += (CYCLES_LCD_MODE0-mode3_cyc);

            if(memory[0xFF41]&0x08)
               set_int(2);

            if(hdma_on)
               do_hdma();                                           
         break;     
         }     
      }
   } else // LCD is off
   {
      off_counter -= cur_cycle;
      if(off_counter <= 0)
      {
         ++memory[0xFF44];
         off_counter += 456;  
      }
       
      if(memory[0xFF44] > 77) //?
         memory[0xFF41] = ((memory[0xFF41]&0xFC)|0x02);
      else
         memory[0xFF41] = memory[0xFF41]&0xFC;
      
      if(memory[0xFF44] == 3 && LCDoff_option && gbc_mode)
      {      
         if(GB->gfx_bit_count == 16)
            fill_gfx_buffers(gfx_pal16[0x7fff]|(gfx_pal16[0x7fff]<<16)); 
         else
            fill_gfx_buffers(gfx_pal32[0x7fff]);           
      }
             
      if(memory[0xFF44] >= 151)
      {
         memory[0xFF44] = 0;
         ++frames;
         
         if(!sgb_multiplayer)
            Check_KBInput(0);
            
         if(!sgb_mask && LCDoff_option)
         {      
            if(sgb_mode)
            {
               fill_gfx_buffers(0UL);
            } else   
            if(gbc_mode)
            {
               if(GB->gfx_bit_count==16)
                  fill_gfx_buffers(gfx_pal16[0x7fff]|(gfx_pal16[0x7fff]<<16)); 
               else
                  fill_gfx_buffers(gfx_pal32[0x7fff]);           
            }
            else
            {
               if(GB->gfx_bit_count==16)
                  fill_gfx_buffers(gfx_pal16[GBC_BGP[0]]|(gfx_pal16[GBC_BGP[0]]<<16)); 
               else
                  fill_gfx_buffers(gfx_pal32[GBC_BGP[0]]);           
            }
         }
         if(!sgb_mask) draw_screen();               
      }
   }

   if(memory[0xFF02]&0x80) // is serial on?
   {
      if(memory[0xFF02]&1) // internal clock
      {     
         cycles_serial -= real_cycle;
         if(cycles_serial <= 0) // end of transmission
         {
            unsigned short data = 0;
               
            if(multiple_gb && (memory_another[0xFF02]&0x80) && !(memory_another[0xFF02]&0x01))
            {
               data_sended = memory[0xFF01];
               data = memory_another[0xFF01];
               
               time_to_switch = 1;
               data_trans = 1;            
            }
            else
               data = serial_function(0); 

            memory[0xFF01] = data; 
            cycles_serial = 0;
            memory[0xFF02] &= 0x7F;
            set_int(8);
            serialbits = 0;
         }
      } else if(/*net_initted ||*/ (connected_device!=DEVICE_NONE && external_clock))
      { 
         cycles_serial -= real_cycle;
         if(cycles_serial <= 0)
         {
          /*  if(net_initted && !link_on)
            {
               check_link_state(1,0);
            }*/
               
          /*  if(net_initted && !data_sent && link_on)
            {
               link_send_data(memory[0xFF01]);
               data_sent = 1;
               //syncing = 1;
            }   */
            ++serialbits;
            if(serialbits == 8) // end of transmission
            {
               unsigned short data = 0xffff; 
               //if(link_on)
                  data = serial_function(1); 
               //else 
               //   data = 0xffff;
                  
               if(data==0xffff)
               {
                  serialbits = 7;
                  //if(link_on)
                  //   cycles_serial += 20;
                  //else
                     cycles_serial += 20;//CYCLES_SERIAL;
               } else
               {
                  memory[0xFF01] = data;  
                  cycles_serial = 0;
                  memory[0xFF02] &= 0x7F;
                  set_int(8);
                  serialbits = 0;
                  //data_sent = 0;     
                  //if(net_initted) link_on = 0;
                  //send_gb_link_state(0,0);
               }
            } else
               cycles_serial += CYCLES_SERIAL;
         }    
      }
   }
   
   if(draw_wait && !sgb_mask)
   {
      draw_wait -= cur_cycle;
      if(draw_wait <= 0)
      {
         draw_wait = 0;
         int r_frame_skip = (speedup&&(speed_up_option&SP_FS9)?9:frame_skip);
         if(memory[0xFF44] < 144 && !(frameskip_counter%(r_frame_skip+1)))
         {
            draw_line_tile();
            sprite_number = 0;
            if((memory[0xFF40] & 0x02) && (video_enable&VID_EN_SPRITE)) 
               draw_sprites();    
         }
      }
   }
   
   if(STAT_wait)
   {
      STAT_wait -= cur_cycle;
      if(STAT_wait <= 0)
      {            
         STAT_wait = 0;
         set_int(2);
      } 
   }

   cycles_sound -= cur_cycle;
   if(cycles_sound <= 0)
   {
      cycles_sound += CYCLES_SOUND;
      
      sound_event();
   }  
 
   if(IME && (memory[0xFF0F]&memory[0xFFFF]))
      handle_interrupt();

   if(EI_count)
   {
      IME = 1;
      EI_count = 0;
   }     
}
}
