/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
    VisualBoyAdvance - a Gameboy and GameboyAdvance emulator
    Copyright (C) 1999-2004 by Forgotten

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
 
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <windows.h>

#include "mem.h"
#include "MBC.h"
#include "debug.h"

#include "GB.h"

void copy_memory(unsigned short to,unsigned short from,int count)
{
   while(count)
   {
      GB->mem_map[to>>12][to&0x0FFF] = GB->mem_map[from>>12][from&0x0FFF];
      ++to; 
      ++from;
      --count;
   }
}

byte *memory_another = NULL;

extern int ramsize[9];

void mem_reset(int gb_number)
{                                          
   bc_select = 0;       
     
   cameraIO = 0;
   RTC_latched = 0;   
   
   rtc.s = 0;
   rtc.m = 0;
   rtc.h = 0;
   rtc.d = 0;
   rtc.control = 0;
   rtc.last_time = time(0);
   rtc.cur_register = 0x08;

   tama_flag = 0;
   tama_time = 0;
   tama_val6 = 0;
   tama_val7 = 0;
   tama_val4 = 0;
   tama_val5 = 0;
   tama_count = 0;
   tama_month = 0;
   tama_change_clock = 0;
   
   HuC3_flag = HUC3_NONE;
   HuC3_RAMvalue = 1;
   
   cart_address = 0;   
}

byte (*readmemory)(unsigned short);

void (*writememory)(unsigned short,byte);

///////////////////////////////////////////////////////
// Saving                                            //
///////////////////////////////////////////////////////

extern char save_directory[300];

bool write_save(int gb_number)
{
   gb_system* old_GB = GB;
   GB = (gb_number==1?GB1:GB2);
   
   char old_directory[400];
   
   GetCurrentDirectory(400,old_directory);
   SetCurrentDirectory(save_directory);
   
   char save_filename[275]; 
   strcpy(save_filename,GB->rom_filename);
   if(gb_number == 1) 
      strcat(save_filename,".sav");
   else
      strcat(save_filename,".sv2");

   FILE* savefile = fopen(save_filename,"wb");
   if(!savefile)
   {
      SetCurrentDirectory(old_directory);
      return false;
   }
   
   if(GB->rom->bankType == MBC7 || GB->rom->bankType == TAMA5)
   {
      if(fwrite(&GB->memory[0xA000],sizeof(byte),256,savefile) < 256)
      {
         fclose(savefile);
         SetCurrentDirectory(old_directory);
         return false;
      }      
   } else
   if(GB->rom->carttype == 6) // MBC2 + battery
   {
      if(fwrite(&GB->memory[0xA000],sizeof(byte),512,savefile) < 512)
      {
         fclose(savefile);
         SetCurrentDirectory(old_directory);
         return false;
      }   
   } else
   if(GB->rom->RAMsize > 2)
   {
      if(fwrite(GB->cartRAM,sizeof(byte),ramsize[GB->rom->RAMsize]*1024,savefile) < ramsize[GB->rom->RAMsize]*1024)
      {
         fclose(savefile);
         SetCurrentDirectory(old_directory);
         return false;
      }
   } else
   {
      if(fwrite(&GB->memory[0xA000],sizeof(byte),ramsize[GB->rom->RAMsize]*1024,savefile) < ramsize[GB->rom->RAMsize]*1024)
      {
         fclose(savefile);
         SetCurrentDirectory(old_directory);
         return false;
      }
   }

   if(GB->rom->RTC || GB->rom->bankType==TAMA5)
   {
      fwrite(&rtc.s, sizeof(int),1,savefile);
      fwrite(&rtc.m, sizeof(int),1,savefile);
      fwrite(&rtc.h, sizeof(int),1,savefile);     
      fwrite(&rtc.d, sizeof(int),1,savefile);   
      fwrite(&rtc.control, sizeof(int),1,savefile); 
      fwrite(&rtc.last_time, sizeof(time_t),1,savefile);               
   }   
   
   if(GB->rom->bankType==TAMA5)
      fwrite(&tama_month, sizeof(int),1,savefile);    
      
   if(GB->rom->bankType == HuC3)
   {
      fwrite(&HuC3_time, sizeof(unsigned int),1,savefile);
      fwrite(&HuC3_last_time, sizeof(time_t),1,savefile); 
      fwrite(&rtc.s, sizeof(int),1,savefile);      
   }
   
   fclose(savefile);
   SetCurrentDirectory(old_directory);
   
   GB = old_GB;   
   return true;
}

bool read_save(int gb_number)
{       
   char old_directory[400];
   
   gb_system* old_GB = GB;
   GB = (gb_number==1?GB1:GB2);
      
   GetCurrentDirectory(400,old_directory);
   SetCurrentDirectory(save_directory);    
     
   char save_filename[275]; 
   strcpy(save_filename,GB->rom_filename);
   if(gb_number==1) 
      strcat(save_filename,".sav");
   else
      strcat(save_filename,".sv2");

   FILE* savefile = fopen(save_filename,"rb");
   if(!savefile) 
   {
      SetCurrentDirectory(old_directory);
      return true;
   }
   
   if(GB->rom->bankType == MBC7 || GB->rom->bankType == TAMA5)
   {
      if(fread(&GB->memory[0xA000],sizeof(byte),256,savefile) < 256)
      {
         fclose(savefile);
         SetCurrentDirectory(old_directory);
         return false;
      }      
   } else
   if(GB->rom->carttype == 6) // MBC2 + battery
   {
      if(fread(&GB->memory[0xA000],sizeof(byte),512,savefile) < 512)
      {
         fclose(savefile);
         SetCurrentDirectory(old_directory);
         return false;
      }   
   } else
   if(GB->rom->RAMsize > 2)
   {
      if(fread(GB->cartRAM,sizeof(byte),ramsize[GB->rom->RAMsize]*1024,savefile) < ramsize[GB->rom->RAMsize]*1024)
      {
         fclose(savefile);
         SetCurrentDirectory(old_directory);
         return false;
      }
   } else
   {
      if(fread(&GB->memory[0xA000],sizeof(byte),ramsize[GB->rom->RAMsize]*1024,savefile) < ramsize[GB->rom->RAMsize]*1024)
      {
         fclose(savefile);
         SetCurrentDirectory(old_directory);
         return false;
      }
   }
   
   if(GB->rom->RTC || GB->rom->bankType==TAMA5)
   {
      fread(&rtc.s, sizeof(int),1,savefile);
      fread(&rtc.m, sizeof(int),1,savefile);
      fread(&rtc.h, sizeof(int),1,savefile);     
      fread(&rtc.d, sizeof(int),1,savefile);   
      fread(&rtc.control, sizeof(int),1,savefile); 
      fread(&rtc.last_time, sizeof(time_t),1,savefile);    
      rtc_latch = rtc;              
   }
   
   if(GB->rom->bankType==TAMA5)
      fread(&tama_month, sizeof(int),1,savefile);       

   if(GB->rom->bankType == HuC3)
   {
      fread(&HuC3_time, sizeof(unsigned int),1,savefile);
      fread(&HuC3_last_time, sizeof(time_t),1,savefile);
      fread(&rtc.s, sizeof(int),1,savefile);            
   }

   fclose(savefile);
   SetCurrentDirectory(old_directory);
   
   GB = old_GB;
   return true;
}

