/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

void gbz80_inc(byte* reg)
{
     asm volatile (    
     "movl $0, %1\n"      // load 0 to NFLAG
     
     "incb %2\n"
          
     "lahf\n"             // get flags
     "andb $1,%%bh\n"     // carry not affected
     "andb $0xFE,%%ah\n"
     "orb %%ah, %%bh \n"
     
     : "=b" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg), "0" (flags)
     : "cc", "ah");  
}

void gbz80_dec(byte* reg)
{
     asm volatile (    
     "movl $1, %1\n"      // load 1 to NFLAG
     
     "decb %2\n"
          
     "lahf\n"             // get flags
     "andb $0x01,%%bh\n"  // carry not affected
     "andb $0xFE,%%ah\n"
     "orb %%ah, %%bh \n"
     
     : "=b" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg), "0" (flags)
     : "cc", "ah"); 
}

void gbz80_rlca()
{
     asm volatile (  
     "xorb %%ah, %%ah\n"       
     "movl $0, %1\n"       // load 0 to NFLAG

     "rolb %2\n"
     
     "rclb %%ah\n"         // only carry affected
   
     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "2" (A)
     : "cc"); 
}

void gbz80_rrca()
{
     asm volatile (  
     "xorb %%ah, %%ah\n"  
     "movl $0, %1\n"       // load 0 to NFLAG
     
     "rorb %2\n"
     
     "rclb %%ah\n"         // only carry affected
         
     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "2" (A)
     : "cc");
}

void gbz80_rla()
{
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG

     "rcrb %%ah\n"         // load carry
     "movb $0, %%ah\n"

     "rclb %2\n"
     
     "rclb %%ah\n"         // only carry affected
          
     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "2" (A), "0" (flags)
     : "cc"); 
}

void gbz80_rra()
{
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG

     "rcrb %%ah\n"         // load carry
     "movb $0, %%ah\n"
     
     "rcrb %2\n"
     
     "rclb %%ah\n"         // only carry affected
     
     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "2" (A), "0" (flags)
     : "cc"); 
}

void gbz80_rlc(byte* reg)
{
     asm volatile (    
     "xorb %%al, %%al\n"      
     "movl $0, %1\n"       // load 0 to NFLAG
          
     "rolb %2\n"           // doesn't affect zero flag
     "rclb %%al\n"         // store carry flag
     "andb $0xff, %2\n"    // check for zero
     "rcrb %%al\n"         // restore carry flag
     
     "lahf\n"              // get flags
     "andb $0xEF, %%ah\n"  // hflag = 0
     
     : "=a" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg)
     : "cc");
}

void gbz80_rrc(byte* reg)
{
     asm volatile (    
     "xorb %%al, %%al\n"      
     "movl $0, %1\n"       // load 0 to NFLAG
          
     "rorb %2\n"           // doesn't affect zero flag
     "rclb %%al\n"         // store carry flag
     "andb $0xff, %2\n"    // check for zero
     "rcrb %%al\n"         // restore carry flag
     
     "lahf\n"              // get flags
     "andb $0xEF, %%ah\n"  // hflag = 0
     
     : "=a" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg)
     : "cc");  
}

void gbz80_rl(byte* reg)
{
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG
     
     "rcrb %%ah\n"           // load carry flag  
     
     "rclb %2\n"           // doesn't affect zero flag
     "rclb %%al\n"         // store carry flag
     "andb $0xff, %2\n"    // check for zero
     "rcrb %%al\n"         // restore carry flag
     
     "lahf\n"              // get flags
     "andb $0xEF, %%ah\n"  // hflag = 0
     
     : "=a" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg), "0" (flags)
     : "cc"); 
}

void gbz80_rr(byte* reg)
{
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG
     
     "rcrb %%ah\n"           // load carry flag  
     
     "rcrb %2\n"           // doesn't affect zero flag
     "rclb %%al\n"         // store carry flag
     "andb $0xff, %2\n"    // check for zero
     "rcrb %%al\n"         // restore carry flag
     
     "lahf\n"              // get flags
     "andb $0xEF, %%ah\n"  // hflag = 0
     
     : "=a" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg), "0" (flags)
     : "cc");
}

void gbz80_sla(byte* reg)
{
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG

     "salb %2\n"
     
     "lahf\n"              // get flags
     "andb $0xEF, %%ah\n"  // hflag = 0
     
     : "=a" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg)
     : "cc");
}

void gbz80_sra(byte* reg)
{
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG

     "sarb %2\n"
     
     "lahf\n"              // get flags
     "andb $0xEF, %%ah\n"  // hflag = 0
     
     : "=a" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg)
     : "cc");
}

void gbz80_srl(byte* reg)
{
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG

     "shrb %2\n"
     
     "lahf\n"              // get flags
     "andb $0xEF, %%ah\n"  // hflag = 0
     
     : "=a" (flags), "=g" (NFLAG), "=q" (*reg)
     : "2" (*reg)
     : "cc");
}

void gbz80_bit(byte reg,int num)
{
     asm volatile (      
     "movb $0x40, %%al\n"    // al = 40h, half carry will be set after rotation
     "movl $0, %1\n"         // load 0 to NFLAG

     "btw %%cx, %%bx\n"      // check bit (load to carry)
     "cmc\n"                 // complement carry
     
     "rclb $7, %%al\n"       // rotate carry to zero flags place
     
     "andb $0x01, %%ah\n"    // doesn't affect carry
     "orb %%al, %%ah\n"      // get stored zero flag
     
     : "=a" (flags), "=g" (NFLAG)
     : "b" (reg), "c" (num), "0" (flags)
     : "cc"); 
}

void gbz80_swap(byte* reg)
{
     asm volatile (      
     "movl $0, %1\n"       // load 0 to NFLAG

     "rolb $4, %2\n"       // swap
     
     "andb $0xff, %2\n"    // check for zero
     "lahf\n"              // get flags
     "andb $0x40, %%ah\n"  // only zflag set/reset

     : "=a" (flags), "=g" (NFLAG), "=q" (*reg) 
     : "2" (*reg)
     : "cc");
}

void gbz80_scf()
{
     asm volatile (      
     "movl $0, %1\n"        // load 0 to NFLAG
     "orw $0x100, %0\n"     // set carry
     "andw $0xEF00, %0\n"   // reset half-carry

     : "=q" (flags), "=g" (NFLAG) 
     : "0" (flags)
     : "cc"); 
}

void gbz80_ccf()
{
     asm volatile (    
     "movl $0, %1\n"      // load 0 to NFLAG
     "btc $8, %0\n"       // complement carry
     "btr $12, %0\n"      // reset half-carry

     : "=q" (flags), "=g" (NFLAG)
     : "0" (flags)
     : "cc"); 
}

void gbz80_cpl()
{
     asm volatile (    
     "movl $1, %1\n"       // load 1 to NFLAG
     "notb %2\n"           // complement A
     "orw $0x1000, %0\n"   // set half-carry

     : "=q" (flags), "=g" (NFLAG), "=q" (A)
     : "2" (A), "0" (flags)
     : "cc");
}

void gbz80_and(byte reg)
{
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG
     
     "andb %3, %2\n"
     
     "lahf\n"              // get flags
   //"andb $0xfe, %%ah\n"  // reset carry
     "orb $16, %%ah\n"     // set half-carry
 
     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "q" (reg), "2" (A)
     : "cc");
}

void gbz80_xor(byte reg)
{
     asm volatile (    
     "movl $0, %1\n"      // load 0 to NFLAG
     
     "xorb %3, %2\n"
     
     "lahf\n"             // get flags
   //"andb $238, %%ah\n"  // reset carry and half-carry

     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "q" (reg), "2" (A)
     : "cc");
}

void gbz80_or(byte reg)
{
     asm volatile (    
     "movl $0, %1\n"      // load 0 to NFLAG
     
     "orb %3, %2\n"
     
     "lahf\n"             // get flags
   //"andb $238, %%ah\n"  // reset carry and half-carry

     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "q" (reg), "2" (A)
     : "cc"); 
}

void gbz80_cp(byte reg)
{
     asm volatile (    
     "movl $1, %1\n"      // load 1 to NFLAG
     
     "cmpb %2, %3\n"
     
     "lahf\n"             // get flags

     : "=a" (flags), "=g" (NFLAG)
     : "q" (reg), "q" (A)
     : "cc");
}

void gbz80_add(byte reg)
{
     asm volatile (    
     "movl $0, %1\n"      // load 0 to NFLAG
     
     "addb %3, %2\n"
     
     "lahf\n"             // get flags

     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "q" (reg), "2" (A)
     : "cc"); 
}

void gbz80_add16(unsigned short* reg, unsigned short reg2)
{
     unsigned short flag = ((((*reg) & 0x0FFF) + (reg2 & 0x0FFF)) > 0x0FFF)<<12;
     
     asm volatile (    
     "movl $0, %1\n"       // load 0 to NFLAG
     "andb $0x40, %%ah\n"  // store zflag
     "movb %%ah, %%al\n"
     "addw %3, %2\n"
     
     "lahf\n"              // get flags
     "andb $0xBF, %%ah\n"  // get stored...
     "orb %%al, %%ah\n"    // ...zflag     

     : "=a" (flags), "=g" (NFLAG), "=r" (*reg)
     : "r" (reg2), "2" (*reg), "0" (flags)
     : "cc"); 
     
     flags &= 0xEF00;
     flags |= flag; 
}

void gbz80_addSP(unsigned short* reg, signed char reg2)
{
     asm volatile (    
     "movl $0, %1\n"      // load 0 to NFLAG

     "cbw\n"              // extend al to ax
     "addw %%ax, %2\n"
     
     "lahf\n"             // get flags
     "btr $14,%%ax\n"     // zflag = 0

     : "=a" (flags), "=g" (NFLAG), "=r" (*reg)
     : "a" (reg2), "2" (*reg)
     : "cc");  
}

void gbz80_adc(byte reg)
{   
     asm volatile (  
     "movl $0, %1\n"      // load 0 to NFLAG
          
     "rcrb %%ah\n"        // load carry flag  
   
     "adcb %3, %2\n"
     
     "lahf\n"             // get flags
     
     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "q" (reg), "0" (flags), "2" (A)
     : "cc");    
}

void gbz80_sub(byte reg)
{
     asm volatile (    
     "movl $1, %1\n"      // load 1 to NFLAG
     
     "subb %3, %2\n"
     
     "lahf\n"             // get flags
     
     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "r" (reg), "2" (A)
     : "cc");    
}

void gbz80_sbc(byte reg)
{
     asm volatile (    
     "movl $1, %1\n"      // load 1 to NFLAG
     
     "rcrb %%ah\n"        // load carry flag  
     
     "sbbb %3, %2\n"
     
     "lahf\n"             // get flags
     
     : "=a" (flags), "=g" (NFLAG), "=q" (A)
     : "q" (reg), "0" (flags), "2" (A)
     : "cc");
}

void gbz80_daa()
{   
     asm volatile (     
     "movb %1, %%al\n" 
     "sahf\n"      
          
     "daa\n"
     
     "lahf\n"             // get flags
     "andb $0xef,%%ah\n"  // hflag = 0    
     "movb %%al,%1\n"

     : "=a" (flags), "=q" (A)
     : "a" (flags), "1" (A)
     : "cc" );   
}

void gbz80_das()
{   
     asm volatile (     
     "movb %1, %%al\n" 
     "sahf\n"      
          
     "das\n"
     
     "lahf\n"             // get flags
     "andb $0xef,%%ah\n"  // hflag = 0    
     "movb %%al,%1\n"
     
     : "=a" (flags), "=q" (A)
     : "a" (flags), "1" (A)
     : "cc" );       
}
