/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/*
    VisualBoyAdvance - a Gameboy and GameboyAdvance emulator
    Copyright (C) 1999-2004 by Forgotten

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "strings.h"

#include "rom.h"
#include "debug.h"
#include "cpu.h"
#include "MBC.h"
#include "mainloop.h"
#include ".\zlib\unzip.h"

#include "GB.h"
#include "mem.h"

int zip_name_option = 0;

void gb_system::crdtype(byte value,byte romsize,byte ramsize)
{
   if((romsize > 8 && romsize < 0x52) || romsize > 0x54) 
      romsize = 0;
   rom->ROMsize = romsize;
   rom->RAMsize = ramsize;
   if(rom->RAMsize >= 8)
      rom->RAMsize = 1;
   rom->RTC = false;
   rom->rumble = false;
   rom->carttype = value;
   readmemory = readmemory_default;
   writememory = writememory_default;
   
   switch(value) 
   {
   case 0x00: rom->battery=false; rom->bankType=ROM; break;  //return "ROM";
   case 0x01: rom->battery=false; rom->bankType=MBC1; writememory = writememory_MBC1; break; //return "MBC1";
   case 0x02: rom->battery=false; rom->bankType=MBC1; writememory = writememory_MBC1; break; //return "MBC1+RAM";
   case 0x03: rom->battery=true; rom->bankType=MBC1; writememory = writememory_MBC1; break;  //return "MBC1+RAM+BATTERY";
   case 0x05: rom->battery=false; rom->bankType=MBC2; writememory = writememory_MBC2; break; //return "MBC2";
   case 0x06: rom->battery=true; rom->bankType=MBC2; writememory = writememory_MBC2; break;  //return "MBC2+BATTERY";
   case 0x08: rom->battery=false; rom->bankType=ROM; break;  //return "ROM+RAM";
   case 0x09: rom->battery=true; rom->bankType=ROM; break;   //return "ROM+RAM+BATTERY";
   case 0x0B: rom->battery=false; rom->bankType=MMM01; writememory = writememory_MMM01; break; //return "MMM01";
   case 0x0C: rom->battery=false; rom->bankType=MMM01; writememory = writememory_MMM01; break; //return "MMM01+RAM";
   case 0x0D: rom->battery=true; rom->bankType=MMM01; writememory = writememory_MMM01; break; //return "MMM01+RAM+BATTERY";
   case 0x0F: rom->battery=true; rom->bankType=MBC3; rom->RTC=true; writememory = writememory_MBC3; readmemory = readmemory_MBC3; break; //return "MBC3+TIMER+BATTERY";
   case 0x10: rom->battery=true; rom->bankType=MBC3; rom->RTC=true; writememory = writememory_MBC3; readmemory = readmemory_MBC3; break; //return "MBC3+TIMER+RAM+BATTERY";
   case 0x11: rom->battery=false; rom->bankType=MBC3; writememory = writememory_MBC3; readmemory = readmemory_MBC3; break; //return "MBC3";
   case 0x12: rom->battery=false; rom->bankType=MBC3; writememory = writememory_MBC3; readmemory = readmemory_MBC3; break; //return "MBC3+RAM";
   case 0x13: rom->battery=true; rom->bankType=MBC3; writememory = writememory_MBC3; readmemory = readmemory_MBC3; break; //return "MBC3+RAM+BATTERY";
   case 0x15: rom->battery=false; rom->bankType=MBC4; break; //return "MBC4";
   case 0x16: rom->battery=false; rom->bankType=MBC4; break; //return "MBC4+RAM"; 
   case 0x17: rom->battery=true; rom->bankType=MBC4;  break; //return "MBC4+RAM+BATTERY";
   case 0x19: rom->battery=false; rom->bankType=MBC5; writememory = writememory_MBC5; break; //return "MBC5";
   case 0x1A: rom->battery=false; rom->bankType=MBC5; writememory = writememory_MBC5; break; //return "MBC5+RAM";
   case 0x1B: rom->battery=true; rom->bankType=MBC5; writememory = writememory_MBC5; break; //return "MBC5+RAM+BATTERY";
   case 0x1C: rom->battery=false; rom->bankType=MBC5; rom->rumble = true; writememory = writememory_MBC5; break; //return "MBC5+RUMBLE";
   case 0x1D: rom->battery=false; rom->bankType=MBC5; rom->rumble = true; writememory = writememory_MBC5; break; //return "MBC5+RUMBLE+RAM";
   case 0x1E: rom->battery=true; rom->bankType=MBC5; rom->rumble = true; writememory = writememory_MBC5; break; //return "MBC5+RUMBLE+RAM+BATTERY";
   case 0x22: rom->battery=true; rom->bankType=MBC7; writememory = writememory_MBC7; readmemory = readmemory_MBC7; break;//Kirby's Tilt'n'Tumble 
   case 0x59: rom->battery=false; rom->bankType=MBC1; writememory = writememory_MBC1; break; //Game Boy Smart Card
   case 0xBE: rom->battery=false; rom->ROMsize++; rom->bankType=MBC5; writememory = writememory_MBC5; break; //Pocket Voice Recorder
   case 0xEA: rom->battery=false; rom->bankType=MBC1; writememory = writememory_MBC1; break; //SONIC5
   case 0xFC: rom->battery=true; rom->bankType=Camera; writememory = writememory_Camera; readmemory = readmemory_Camera;  break; //return "POCKET CAMERA";
   case 0xFD: rom->battery=true; rom->bankType=TAMA5; writememory = writememory_TAMA5; readmemory = readmemory_TAMA5; break; //return "Bandai TAMA5";
   case 0xFE: rom->battery=true; rom->bankType=HuC3; writememory = writememory_HuC3; readmemory = readmemory_HuC3; break; //return "Hudson HuC-3";
   case 0xFF: rom->battery=true; rom->bankType=HuC1; writememory = writememory_MBC1; break; //return "Hudson HuC-1+RAM+BATTERY";
   default: rom->battery=false; rom->bankType=Unknown; debug_print(str_table[ERROR_ROM_TYPE]); break; //return "Unidentified";
   }
}

int gb_system::process_rom_info(byte *rominfo)
{
   int addr = 0;
   byte data;
   for(;addr<=14; ++addr)
      rom->name[addr] = rominfo[addr];
      
   int thisromsize = 0;

   if(rominfo[addr] == 0x80) 
      rom->CGB = 1; 
   else if(rominfo[addr] == 0xC0)
      rom->CGB = 2; // gbc only
   else
      rom->CGB = 0;
   // modes changed when reset
          
   ++addr; rom->newlic = (rominfo[addr]<<8);
   ++addr; rom->newlic |= rominfo[addr];
   ++addr; 
   if(rominfo[addr] == 0x03) 
      rom->SGB = 1; 
   else 
      rom->SGB = 0;
      
   ++addr; crdtype(rominfo[addr],data=rominfo[addr+1],rominfo[addr+2]);
      
   if(!strcmp(rom->name,"GB SMART CARD"))
   {
      data = 0;
      rom->ROMsize = 0;    
   }     
   
   // BHGOS MultiCart
   if(!strcmp(rom->name,"MultiCart"))
   { 
      readmemory = readmemory_default;
      writememory = writememory_default;
      rom->RAMsize = 3;
      rom->ROMsize = 2;
   }
   
   switch(data)
   {
      case 0: thisromsize=2; break;
      case 1: thisromsize=4; break;
      case 2: thisromsize=8; break;  
      case 3: thisromsize=16; break; 
      case 4: thisromsize=32; break;  
      case 5: thisromsize=64; break; 
      case 6: thisromsize=128; break;
      case 7: thisromsize=256; break;
      case 8: thisromsize=512; break;
      case 0x52: thisromsize=72; break;
      case 0x53: thisromsize=80; break;
      case 0x54: thisromsize=96; break;
      default: thisromsize=2; debug_print(str_table[ERROR_ROM_SIZE]); break;
   }

   char ball_name[16] = { 0x42,0x61,0x6C,0x6C,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x00,0x80 };    
   // Ball (Bung)(PD)[C] and Capman (Lik-Sang)(PD)[C] and Fix & Foxi [C][t1]
   if(!strcmp(rom->name,ball_name) || strstr(rom->name,"CAPMAN") || !strcmp(rom->name,"LUPO +3HI"))
   {
      writememory = writememory_default;
   }
   else    
   //Bugs Bunny - Crazy Castle 3 (J)[C][t2]
   if(!strcmp(rom->name,"BUGS CC3 CRACK"))
   {
      writememory = writememory_default; 
      rom->RAMsize=1;
   }
   else
   //Pulsar (Freedom GB Contest 2001)(PD)[C]
   if(strstr(rom->name,"PULSAR"))
   {
      rom->bankType=MBC5;
   }
   else
   //Pokemon Red-Blue 2-in-1 (Unl)[S][h1]
   if(!strcmp(rom->name,"POKEMON RED") && rom->ROMsize == 6)
   {
      rom->ROMsize = 7;
      writememory = writememory_poke;
   }
   else   
   // SGB Pack
   if(!strcmp(rom->name,"SGBPACK"))
   {
      thisromsize = 2;
      rom->ROMsize = 0;
   }
   else        
   // Dragon Ball Z Goku (Chinese)
   if(!strcmp(rom->name,"GB DBZ GOKOU"))  
   {
      rom->carttype = 3;
      rom->battery=true; rom->bankType=MBC1; writememory = writememory_MBC1;
   }
   else       
   // Bokujou Monogatari 3 Chinese
   if(!strcmp(rom->name,"BOKUMONOGB3BWAJ") || !strcmp(rom->name,"BOYGIRLD640BWAJ"))
   {
      thisromsize = 128;
      rom->ROMsize = 6;
   }
   else      
   // Monsters GO!GO!GO!!
   if(!strcmp(rom->name,"POCKET MONSTER"))
   {
      thisromsize = 32;
      rom->ROMsize = 4;
   }
   else      
   // Sonic 3D Blast 5
   if(strstr(rom->name,"SONIC5"))
   {
      thisromsize = 16;
      rom->ROMsize = 3;
      rom->RAMsize = 0;
   }
   else   
   // Rockman 8
   if(!strcmp(rom->name,"ROCKMAN 99"))
      writememory = writememory_Rockman8;
   else
   // Collection Carts
   if(!strcmp(rom->name,"BOMCOL") || !strcmp(rom->name,"GENCOL") || strstr(rom->name,"MOMOCOL") || strstr(rom->name,"SUPERCHINESE 12"))
   {
      writememory = writememory_BC;
      if(strstr(rom->name,"MOMOCOL2"))
         writememory = writememory_MMM01;         
   }
   else
   if(!strcmp(rom->name,"MORTALKOMBATI&I"))
   {
      writememory = writememory_MK12;
   }
   else
   // Gameboy Camera
   if(!strcmp(rom->name,"GAMEBOYCAMERA"))
   {
      rom->bankType = Camera;
      thisromsize = 64;
      rom->ROMsize = 5;
      rom->RAMsize = 4;
      writememory = writememory_Camera;
      readmemory = readmemory_Camera;
   }
      
   addr+=3; rom->destcode = rominfo[addr];
   ++addr; rom->lic = rominfo[addr];
   ++addr; rom->version = rominfo[addr];
   ++addr; rom->complement = rominfo[addr];
   ++addr; rom->checksum=(rominfo[addr]<<8);
   ++addr; rom->checksum|=rominfo[addr];

   // Digimon 3 saving
   if(!strcmp(rom->name,"DIGIMON") && rom->checksum == 0xE11B)
      rom->battery = true;
   else   
   // Joust & Defender (U)[C][t1]
   if(!strcmp(rom->name,"DEFENDER/JOUST") && rom->checksum == 0xB110)
      rom->RAMsize = 1;
           
   // check complement
   byte cmpl=0;
   for(addr=0;addr<=0x19;++addr) 
      cmpl+=rominfo[addr];
   cmpl+=25; rom->complementok = !cmpl;
   // done processing info...

   return thisromsize;
}

int get_size(int real_size) // return good size to use
{
   if(real_size < 64*1024)
      return 64*1024;
   if(real_size < 128*1024)
      return 128*1024;
   else
   if(real_size < 256*1024)
      return 256*1024;
   else
   if(real_size < 512*1024)
      return 512*1024;
   else
   if(real_size < 1024*1024)
      return 1024*1024;
   else
   if(real_size < 2048*1024)
      return 2048*1024;   
   else
   if(real_size < 4096*1024)
      return 4096*1024;
   else
   if(real_size < 8192*1024)
      return 8192*1024;
   else
      return real_size;                                 
}

bool gb_system::loadrom_zip(char* filename)
{
   memset(rom_filename,0,ROM_FILENAME_SIZE);  
   if(zip_name_option == USE_ZIPFILE_NAME)
   {
      char* temp = strrchr(filename,'\\'); // find last '\'
      if(temp == NULL)
         temp = filename;
      else
         temp += 1;
      
      strncpy(rom_filename,temp,strlen(temp)-4);
   }   
   byte rominfo[30];  
   
   romloaded = false;
   unzFile unz = unzOpen(filename);
   if(unz == NULL)
   {
      debug_print(str_table[ERROR_OPEN_FILE]);
      return false;
   }
   
   int r = unzGoToFirstFile(unz);
    
   if(r != UNZ_OK) 
   {
      unzClose(unz);
      debug_print(str_table[ERROR_FIND_FILE]);
      return false;
   }
   
  bool found = false;
    
  unz_file_info info;
  char buffer_file_name[300];
  
   while(true) 
   {
      r = unzGetCurrentFileInfo(unz,&info,buffer_file_name,sizeof(buffer_file_name),NULL,0,NULL,0);
      
      if(r != UNZ_OK) 
      {
         unzClose(unz);
         debug_print(str_table[ERROR_BAD_ZIP]);
         return false;
      }

      if(strstr(buffer_file_name,".gb") || strstr(buffer_file_name,".GB") || strstr(buffer_file_name,".SGB") || strstr(buffer_file_name,".sgb")) 
      {
         if(zip_name_option == USE_ROMFILE_NAME)
         {
            if(strstr(buffer_file_name,".gbc") || strstr(buffer_file_name,".GBC") || strstr(buffer_file_name,".SGB") || strstr(buffer_file_name,".sgb"))
               strncpy(rom_filename,buffer_file_name,strlen(buffer_file_name)-4);
            else
               strncpy(rom_filename,buffer_file_name,strlen(buffer_file_name)-3);
         }
        
         found = true;
         break;
      }
    
      r = unzGoToNextFile(unz);
      
      if(r != UNZ_OK)
         break;
   }

   if(!found) 
   {
      unzClose(unz);
      debug_print(str_table[ERROR_FIND_ROM]);
      return false;
   }
  
   int size = info.uncompressed_size;
    
   r = unzOpenCurrentFile(unz);

   if(r != UNZ_OK) 
   {
      unzClose(unz);
      debug_print(str_table[ERROR_OPEN_ZIP_FILE]);
      return false;
   }
   
   if(cartridge != NULL) 
   { 
      delete [] cartridge; 
      cartridge = NULL; 
   }
      
   cartridge = new byte[get_size(size)];
   if(cartridge == NULL) 
   { 
      unzCloseCurrentFile(unz);
      unzClose(unz);         
      debug_print(str_table[ERROR_MEMORY]); 
      return false; 
   }

   if(size != unzReadCurrentFile(unz,cartridge,size)) 
   {
      debug_print(str_table[ERROR_READ_FILE]);
      delete [] cartridge;
      cartridge = NULL;
      return false;
   }
   memcpy(rominfo,cartridge+0x0134,0x1C);
  
   unzCloseCurrentFile(unz);
   unzClose(unz);
  
   process_rom_info(rominfo);
   if(!strcmp(rom->name,"TETRIS") && size > 32768 && rom->ROMsize==0)
   {
      writememory = writememory_MBC1;
      rom->ROMsize = 2;
   } else
   if(!strcmp(rom->name,"\0") && size > 32768 && rom->ROMsize==0)
   {
      writememory = writememory_8in1;
      rom->ROMsize = 4;
   } else
   if(strstr(rom->name,"TUWAMONO") && size == 524288) // Joryu Janshi Ni Chousen (J)[C] (bad)
   {
      debug_print("Bad dump!");
      rom->ROMsize--;
   }  else      
   if(!strcmp(rom->name,"SGBPACK") && size > 32768)
   {
      rom->ROMsize = 6;
      writememory = writememory_poke;
   } else     
   if(size == 262144 && rom->ROMsize==4)
      rom->ROMsize--;
      
   romloaded = true;

   return true;
}

bool gb_system::load_rom(char* filename)
{     
   if(strstr(filename,".zip") || strstr(filename,".ZIP"))
      return loadrom_zip(filename);
        
   struct stat file_stat;

   memset(rom_filename,0,ROM_FILENAME_SIZE);
   byte rominfo[30];

   romloaded = false;
   FILE* romfile = fopen(filename,"rb");
   if(!romfile) 
   { 
      debug_print(str_table[ERROR_OPEN_FILE]); 
      return false; 
   }
  
   // Go to start of ROM info area
   if(fseek(romfile,0x0134,SEEK_SET))
   { 
      debug_print(str_table[ERROR_FSEEK]); 
      fclose(romfile);
      return false; 
   }
               
   // Read internal ROM information area...
   if(fread(rominfo,1,0x1C,romfile)<0x1C)
   { 
      debug_print(str_table[ERROR_READ_ROM_INFO]); 
      fclose(romfile);
      return false; 
   }      
      
   // ...process it...
   int thisromsize = process_rom_info(rominfo);
   
   int file_size = 0;
   
   if(stat(filename,&file_stat) == 0)
   {
      file_size = file_stat.st_size;
      
      // Captain Knick Knack
      if(!strcmp(rom->name,"TETRIS") && file_size > 32768 && rom->ROMsize==0)
      {
         writememory = writememory_MBC1;
         rom->ROMsize = 2;
      } else
      if(!strcmp(rom->name,"\0") && file_size > 32768 && rom->ROMsize==0)
      {
         writememory = writememory_8in1;
         rom->ROMsize = 4;
      } else
      if(strstr(rom->name,"TUWAMONO") && file_size == 524288) // Joryu Janshi Ni Chousen (J)[C]
      {
         debug_print("Bad dump!");
         rom->ROMsize--;
      } else      
      if(!strcmp(rom->name,"SGBPACK") && file_size > 32768)
      {
         rom->ROMsize = 6;
         writememory = writememory_poke;
      }
      if(file_size == 262144 && rom->ROMsize==4)
         rom->ROMsize--;
   }
   else // failed ?
      file_size = 16384 * thisromsize;
   
   // ...now load the rom into memory
   rewind(romfile);  // go to beginning
   
   if(cartridge != NULL) 
   { 
      delete [] cartridge; 
      cartridge=NULL; 
   }
   cartridge = new byte[get_size(file_size)];
   if(cartridge == NULL) 
   { 
      debug_print(str_table[ERROR_MEMORY]); 
      fclose(romfile);
      return false; 
   }
   if(fread(cartridge,1,file_size,romfile)<file_size)
   { 
      if(rom->ROMsize)
         debug_print(str_table[ERROR_READ_ROM_TO_MEMORY]); 
      //don't care...
   }    

   // Get filename
   char* temp = strrchr(filename,'\\'); // find last '\'
   if(temp == NULL)
      temp = filename;
   else
      temp += 1;
      
   char* temp2 = strrchr(temp,'.'); // find last '.' 
   if(temp2 == NULL) // no extension
      strncpy(rom_filename,temp,strlen(temp));
   else
      strncpy(rom_filename,temp,strlen(temp)-strlen(temp2));

   romloaded = true;
   fclose(romfile);
     
   return true;
}


